import sys
import tornado.websocket
import tornado.ioloop
import tornado.httpserver
import tornado.web
import json
import time
import math
import os
from global_vars import imu
import binascii
import global_vars as gl


# note: version string update should follow the updating rule
server_version = '1.1.1'

callback_rate = 50
class WSHandler(tornado.websocket.WebSocketHandler):
    count = 0
    magProgress = 0

    def open(self):
        self.callback = tornado.ioloop.PeriodicCallback(self.send_data, callback_rate)
        self.callback.start()
        self.callback2 = tornado.ioloop.PeriodicCallback(self.detect_status, 500)
        self.callback2.start()

    def detect_status(self):        
        if not imu.read(200):
            self.write_message(json.dumps({ "messageType": "queryResponse","data": {"packetType": "DeviceStatus","packet": { "returnStatus":1}}}))

    def send_data(self):
        if not imu.device_id:            
            self.write_message(json.dumps({ "messageType": "queryResponse","data": {"packetType": "DeviceStatus","packet": { "returnStatus":1}}}))
            time.sleep(1)
        if not imu.paused:
            d = imu.get_latest()
            self.write_message(json.dumps({ 'messageType' : 'event',  'data' : { 'newOutput' : d }}))               
        else:
            return False

    def on_message(self, message):
        global imu

        message = json.loads(message)
        # Except for a few exceptions stop the automatic message transmission if a message is received
        if message['messageType'] != 'serverStatus' and list(message['data'].keys())[0] != 'startLog' and list(message['data'].keys())[0] != 'stopLog':
            self.callback.stop()
            imu.pause()
        if message['messageType'] == 'serverStatus':
            if imu.logging:
                fileName = imu.logger.user['fileName']
            else:
                fileName = ''
            
            # Load the basic openimu.json(IMU application)
            #with open('app_config/IMU/openimu.json') as json_data:
            #    imu.imu_properties = json.load(json_data)
            application_type = bytes.decode(imu.openimu_get_user_app_id())           
            # application_type = imu.device_id
            for idx, item in enumerate(gl.app_str):
                if item in application_type:
                    folder_path = gl.string_folder_path.replace('APP_TYP',item)
                    break   
                            
            # load application type from firmware 
            try:
                if imu.paused == 1 and not imu.openimu_get_user_app_id() == None: 
                    with open(folder_path) as json_data:
                            imu.imu_properties = json.load(json_data)
                    js_version_str = imu.imu_properties['app_version'].split(' ')[2]
                    imu.openimu_version_compare(application_type,js_version_str)
                    self.write_message(json.dumps({ 'messageType' : 'serverStatus', 'data' : { 'serverVersion' : server_version, 'serverUpdateRate' : callback_rate,  'packetType' : imu.packet_type,
                                                                                                'deviceProperties' : imu.imu_properties, 'deviceId' : imu.device_id, 'logging' : imu.logging, 'fileName' : fileName }}))
                else:
                    self.write_message(json.dumps({ "messageType": "queryResponse","data": {"packetType": "DeviceStatus","packet": { "returnStatus":2}}}))
                    imu.pause()
            except Exception as e:
                # print(e)                 
                self.write_message(json.dumps({ "messageType": "queryResponse","data": {"packetType": "DeviceStatus","packet": { "returnStatus":2}}}))
                imu.pause()
        elif message['messageType'] == 'requestAction':
            if list(message['data'].keys())[0] == 'gA':                
                data = imu.openimu_get_all_param()

                # data[7]['value'] = data[3]['value'].strip(b'\x00'.decode())
                time.sleep(0.2)
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "gA" : data }}))
                # print('requesting ok ---------------{0}'.format(self.count))                
            elif list(message['data'].keys())[0] == 'uP':
                data = imu.openimu_update_param(message['data']['uP']['paramId'], message['data']['uP']['value'])
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "uP" : data }}))
            elif list(message['data'].keys())[0] == 'sC':
                imu.openimu_save_config()
                time.sleep(0.5)
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "sC" : {} }}))
            # added by dave, for connect page to show version
            elif list(message['data'].keys())[0] == 'gV':
                data = imu.openimu_get_user_app_id()
                self.write_message(json.dumps({ "messageType" : "completeAction", "data" : { "gV" : str(data) }}))
            elif list(message['data'].keys())[0] == 'startStream':                                
                imu.connect()
                self.callback.start()  
                self.callback2.stop()
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "startStream" : {} }}))
            elif list(message['data'].keys())[0] == 'stopStream':
                imu.pause()
                self.callback2.start()
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "stopStream" : {} }}))
            elif list(message['data'].keys())[0] == 'startLog' and imu.logging == 0: 
                data = message['data']['startLog']
                imu.start_log(data)
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "logfile" : imu.logger.name }}))
            elif list(message['data'].keys())[0] == 'stopLog' and imu.logging == 1: 
                imu.stop_log()                
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "logfile" : '' }}))
            # added by Dave, app download page
            elif list(message['data'].keys())[0] == 'upgradeFramework':
                fileName = message['data']['upgradeFramework']
                if imu.openimu_upgrade_fw_prepare(fileName):
                    while not imu.openimu_finish_upgrade_fw():
                        imu.openimu_upgrade_fw(fileName)
                        self.write_message(json.dumps({ "messageType" : "processAction", "data" : { "addr" : imu.addr, "fs_len": imu.fs_len }}))
                    imu.openimu_start_app()
                self.write_message(json.dumps({ "messageType" : "completeAction", "data" : { "upgradeFramework" : fileName }}))
        elif message['messageType'] == 'magAction':
            if (list (message['data'].values())[0] == 'start'):
                imu.magneticAlignCmd('start')
                self.magProgress = 1
                # print ('mag align started')
                self.write_message(json.dumps({"messageType": "magAction", "data": {"start": {}}}))
            elif (list(message['data'].values())[0] == 'abort'):
                time.sleep(2)
                imu.magneticAlignCmd('abort')
                self.magProgress = 0
                # print ('mag align aborted')
                self.write_message(json.dumps({"messageType": "magAction", "data": {"abort": {}}}))
                return

            elif (list(message['data'].values())[0] == 'status'):
                # status = openIMUMagneticAlign.status()
                if (self.magProgress == 1):
                    status = imu.magneticAlignCmd('status')

                    if status == 1:
                        time.sleep(1)
                        storedValue = imu.magneticAlignCmd('stored')
                        self.write_message(
                            json.dumps({"messageType": "magAction", "data": {"status": "complete", "value": storedValue}}))

                        return


                    self.write_message(json.dumps({"messageType": "magAction", "data": {"status": "incomplete"}}))

            elif (list(message['data'].values())[0] == 'save'):
                imu.magneticAlignCmd('save')
                time.sleep(1)
                self.magProgress = 0
                data = imu.openimu_get_all_param()
                self.write_message(json.dumps({"messageType": "magAction", "data": {"status": "saved","value":data}}))
                return




        # OLD CODE REVIEW FOR DELETION
        elif  0 and message['messageType'] == 'requestAction':
            # Send and receive file list from local server
            if list(message['data'].keys())[0] == 'listFiles':
                logfiles = [f for f in os.listdir('data') if os.path.isfile(os.path.join('data', f)) and f.endswith(".csv")]
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "listFiles" : logfiles }}))
            elif list(message['data'].keys())[0] == 'loadFile':
                print(message['data']['loadFile']['graph_id'])
                f = open("data/" + message['data']['loadFile']['graph_id'],"r")
                self.write_message(json.dumps({ "messageType" : "requestAction", "data" : { "loadFile" :  f.read() }}))

    def on_close(self): 
        if(imu.logging == 1):
            imu.stop_log() 
        imu.pause()
        self.callback.stop()
        self.callback2.stop()
        time.sleep(1.2)
        
        
        # try:
        #     os._exit(0)
        # except:
        #    print('Program is off.')

        return False

    def check_origin(self, origin):
        return True
 
if __name__ == "__main__":
    # Create IMU
    print("server_version:",server_version)
    try: 
        imu.find_device()    
        # Set up Websocket server on Port 8000
        # Port can be changed
        application = tornado.web.Application([(r'/', WSHandler)])
        http_server = tornado.httpserver.HTTPServer(application)
        http_server.listen(8000)        
        
        tornado.ioloop.IOLoop.instance().start()
    
    except KeyboardInterrupt:  # response for KeyboardInterrupt such as Ctrl+C
        print('User stop this program by KeyboardInterrupt! File:[{0}], Line:[{1}]'.format(__file__, sys._getframe().f_lineno))
        os._exit(1)
    except Exception as e:
        print(e)    

    



"""
Python command line interface
Date: 07-05-2018
Author:
"""

import serial
import string
import time
import json
import time
import threading
import math
import os
import sys
import tornado.websocket
import tornado.ioloop
import tornado.httpserver
import tornado.web
import psutil

from imu_input_packet import InputPacket
from server import WSHandler
from global_vars import imu
from os import getpid

class OpenIMU_CLI:
    def __init__(self):
        '''Initialize command line interface
        '''
        self.command = "help"         # set default command to help
        self.input_string = []        # define input_string
        self.cli_properties = [] 
        self.current_command =[]
        self.baud_rate = 0
        self.accel_lpf = 0
        self.rate_lpf = 0
        self.orien = 0
        self.thread = threading.Thread(target=self.start_tornado)
        self.http_server = None
        self.http_server_running = False

    def help_handler(self):
        print("Usage: ")
        for x in self.cli_properties:
            print(x['name'] + " : " + x['description']) 
        
        return True

    def connect_handler(self):
        '''connect command is used to find OpenIMU device
        '''
        imu.find_device()

        return True

    def upgrade_handler(self):
        '''upgrade command is used for firmware upgrade and followed by file name
        '''
        input_args = len(self.input_string)
        if (input_args == 1):
            print("Usage:")
            print("upgrade file_name")
        else:
            file_name = self.input_string[1]
            imu.ws = False
            if file_name is not " ":
                if imu.openimu_upgrade_fw_prepare(file_name) == True:
                    while not imu.openimu_finish_upgrade_fw():
                        imu.openimu_upgrade_fw(file_name)
                    imu.openimu_start_app()
                else:
                    print("Error: file doesn't exist")  
        return True

    def record_handler(self):
        '''record command is used to save the outputs into local machine
        '''
        imu.ws = False
        imu.start_log()
        return True

    def stop_handler(self):
        '''record command is used to save the outputs into local machine
        '''
        if imu.logging == 1 or self.http_server_running == True:
            if imu.logging == 1:
                imu.stop_log()
            if self.http_server_running ==True:
                self.server_stop()
        else:
            print("no ongoing recording or server not running now!")
        time.sleep(2)
        return True

    def run_handler(self):
        '''used by customers
        '''
        print("Running...")
        self.server_start_handler()
        return True

    def get_handler(self):
        '''get configuration or data from OpenIMU device
        '''
        input_args = len(self.input_string)
        param_properties = imu.imu_properties['userConfiguration']
 
        if (input_args == 1):
            print("Usage: get [options]")
            print("Option: ")
            i = 2
            while i < len(param_properties):
                print(param_properties[i]['argument'])
                i += 1
            return True
        else:
            i = 2
            while i < len(param_properties):
                x = param_properties[i]
                if (x['argument'] == self.input_string[1]):
                    break
                i += 1
                if (i == len(param_properties)):
                    print("Usage: get [options]")
                    print("Option: ")
                    i = 2
                    while i < len(param_properties):
                        print(param_properties[i]['argument'])
                        i += 1
                    return True
        
        param = imu.openimu_get_param(x['paramId'])
        return True
   
    def set_param(self, command, value):
        param_properties = imu.imu_properties['userConfiguration']
        i = 2
        while i < len(param_properties):
            x = param_properties[i] 
            if (x['argument'] == command):
               break
            i += 1

        if i < len(param_properties):
            if command == "type" or command == "orien":
                imu.openimu_update_param(x['paramId'], value)
            else:
                imu.openimu_update_param(x['paramId'], int(value))
        else:
            print("Usage: set [options] <value>") 
            
        return True
     
    def set_handler(self):
        '''set parameters' values
        '''
        input_args = len(self.input_string)
        param_properties = imu.imu_properties['userConfiguration']

        if input_args == 1:
            print("Usage: set <options> <values>")
            i = 2
            while i < len(param_properties):
                x = param_properties[i] 
                print(x['argument'])
                i += 1
            return True
        else:
            i = 2
            while i < len(param_properties):
                x = param_properties[i] 
                if (x['argument'] == self.input_string[1]):
                    break
                i += 1

        if input_args == 2:
            if i == len(param_properties):
                print("Usage: set <options> <values>")
                i = 2
                while i < len(param_properties):
                    x = param_properties[i] 
                    print(x['argument'])
                    i += 1
            else:
                print("Usage: set " + x['argument'] + " <values>")
                print("values: ") 
                print(x['options'])
            return True

        if ((x['type'] == "char8" and self.input_string[2] not in x['options']) or 
            (x['type'] == "int64" and int(self.input_string[2]) not in x['options'])):
                print("Usage: set " + x['argument'] + " <values>")
                print("values: ") 
                print(x['options'])
                return True
        else:
            self.set_param(self.input_string[1], self.input_string[2])

        return True

    def save_handler(self):
        C = InputPacket(imu.imu_properties, 'sC')
        imu.write(C.bytes)
        return True
   
    def start_tornado(self):
        application = tornado.web.Application([(r'/', WSHandler)])
        self.http_server = tornado.httpserver.HTTPServer(application)
        if sys.version_info[0] > 2:
            import asyncio
            asyncio.set_event_loop(asyncio.new_event_loop())
        self.http_server.listen(8000)
        tornado.ioloop.IOLoop.instance().start()
         
    def server_start_handler(self):
        imu.ws = True
        self.thread.start()
        self.http_server_running = True
        return True

    def server_stop(self):
        print("server stops")
        if self.http_server is not None:
           self.http_server.stop()
        imu.ws = False
        self.http_server_running = False
        pid = getpid()
        p = psutil.Process(pid)
        p.kill() 

    def command_handler(self):
        '''main routine to handle input command
        '''       
        self.cli_properties = imu.imu_properties['CLICommands']

        while True:
            if sys.version_info[0] < 3:
                token = raw_input(">>")
            else:
                token = input(">>")
            self.input_string = token.split(" ")
             
            if token.strip() == 'exit':
                if self.http_server_running == True:
                    self.http_server_running = False
                    self.server_stop()
                else:
                    break
            if self.http_server_running == True and token.strip() != 'stop':
                print("server is on-going, please stop it")
                continue

            for x in self.cli_properties:
                if x['name'] == self.input_string[0]:
                    self.current_command = x
                    eval('self.%s()'%(x['function']))
                    break
            else:
                self.help_handler()    

        return True 

if __name__ == "__main__":
    imu.find_device()
    cli = OpenIMU_CLI()
    cli.command_handler()



"""
Driver for Aceinna OpenIMU Series Products
Based on PySerial https://github.com/pyserial/pyserial
Created on 2019-08-30
@author: m5horton
"""


"""
command line: python bootloader.py <binary image name>


WS Master Connection 
connect         - finds device, gets device_id/odr_setting, and loops
                - run this in thread otherwise blocking
disconnect      - ends loop

Device Discovery
find_device     - entry point to find a serial connected IMU
find_ports
autobaud

Logging
start_log
stop_log

Control EEPROM Config Fields
get_fields
set_fields
read_fields
write_fields

Bootloader Functions
upgrade_fw      - entry point to flash a serial connected IMU
start_bootloader
write_block
start_app

Syncing 
sync            - trys to sync to a unit continuously transmitting
set_quiet       - sets unit to stop continuous transmission (stream_mode = 0)
restore_odr     - restores unit to whatever odr_setting is

Data Functions
get_latest
get_packet
get_id_str
get_bit_status
parse_packet
calc_crc

Serial          - a tiny layer on top of Pyserial to handle exceptions as means of device detection
open
close
read
write
reset_buffer

Ping
ping_test

"""

import serial
import math
import string
import quat
import time
import sys
import file_storage
import collections
import glob

class OpenIMU:
    def __init__(self, ws=False):
        '''Initialize and then start ports search and autobaud process
        '''
        self.ws = ws                # set to true if being run as a thread in a websocket server
        self.ser = None             # the active UART
        self.synced = 0             # synced status in streaming mode
        self.stream_mode = 0        # 0 = polled, 1 = streaming, commanded by set_quiet and restore_odr
        self.device_id = 0          # unit's id str
        self.connected = 0          # imu is successfully connected to a com port, kind of redundant with device_id property
        self.odr_setting = 0        # value of the output data rate EEPROM setting
        self.logging = 0            # logging on or off
        self.logger = None          # the file logger instance
        self.packet_size = 0        # expected size of packet 
        self.packet_type = 0        # expected type of packet
        self.elapsed_time_sec = 0   # an accurate estimate of elapsed time in ODR mode using IMU timer data
        self.data = {}              # placeholder imu measurements of last converted packeted
        self.filename = sys.argv[1]
        self.ID = ''
        self.port = 'none'
        self.boot = 0
		
    def find_device(self):
        ''' Finds active ports and then autobauds units, repeats every 2 seconds
        '''
        while not self.autobaud(self.find_ports()):
            time.sleep(2)

    def find_ports(self):
        ''' Lists serial port names. Code from
            https://stackoverflow.com/questions/12090503/listing-available-com-ports-with-python
            Successfully tested on Windows 8.1 x64, Windows 10 x64, Mac OS X 10.9.x / 10.10.x / 10.11.x and Ubuntu 14.04 / 14.10 / 15.04 / 15.10 with both Python 2 and Python 3.
            :raises EnvironmentError:
                On unsupported or unknown platforms
            :returns:
                A list of the serial ports available on the system
        '''
        print('scanning ports')
        if sys.platform.startswith('win'):
            ports = ['COM%s' % (i + 1) for i in range(256)]
        elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
            # this excludes your current terminal "/dev/tty"
            ports = glob.glob('/dev/tty[A-Za-z]*')
        elif sys.platform.startswith('darwin'):
            ports = glob.glob('/dev/tty.*')
        else:
            raise EnvironmentError('Unsupported platform')

        result = []
        for port in ports:
            try:
                print('Trying: ' + port)
                s = serial.Serial(port)
                s.close()
                result.append(port)
            except (OSError, serial.SerialException):
                pass
        return result

    def autobaud(self, ports):
        '''Autobauds unit - first check for stream_mode / continuous data, then check by polling unit
           :returns: 
                true when successful
        ''' 
        
        for port in ports:
            for baud in [115200, 57600, 38400]:
                self.open(port, baud)
                # sync() works for stream mode
                self.sync()
                if self.stream_mode:
                    print('Connected Stream Mode ' + '{0:d}'.format(baud) + '  ' + port)
                    break
                else:
                    self.ser.close()
            
            # stream mode not found for port, check port by polling
            if self.stream_mode == 0:  
                for baud in [115200, 57600, 38400]:
                    self.open(port, baud)
                    self.device_id = self.get_id_str()
                    if self.device_id:
                        print('Connected Polled Mode ' + '{0:d}'.format(baud))
                        odr = self.read_fields([0x0001], 1)
                        if odr:
                            print('Saved ODR: ' + '{0:d}'.format(odr[0][1]))
                            self.odr_setting = odr[0][1]
                        self.connected = 1
                        self.port = port
                        return True; 
                    else:
                        self.close()

            # in stream stream mode worked, get odr field and id str
            else:
                odr = self.read_fields([0x0001], 1)
                if odr:
                    print('Current ODR: ' + '{0:d}'.format(odr[0][1]))
                    self.odr_setting = odr[0][1]
                    self.device_id = self.get_id_str()  # read device string
                    self.restore_odr()
                    self.connected = 1                  # a valid connection exists to unit
                    return True
                else:
                    print('failed to get id string')
                    return False
        
        return False

    def get_latest(self):
        '''Get latest converted IMU readings in converted units
            :returns:
                data object or error message for web socket server to pass to app
        '''
        if self.stream_mode == 1:
            return self.data
        else: 
            return { 'error' : 'not streaming' }
    
    def start_log(self, data):
        '''Creates file or cloud logger.  Autostarts log activity if ws (websocket) set to false
        '''
        self.logging = 1
        self.logger = file_storage.LogIMU380Data(self,data)
        if self.ws == False and self.odr_setting != 0:
            self.connect()
    
    def stop_log(self):
        '''Stops file or cloud logger
        '''
        self.logging = 0
        self.logger.close()
        self.logger = None

    def ping_test(self):
        '''Executes ping test.  Not currently used
            :returns:
                True is successful
        '''
        self.stream_mode = 0      
        C = [0x55, 0x55, 0x50, 0x4B, 0x00]      # 0x55504B00
        crc = self.calc_crc(C[2:4] + [0x00])    # for some reason must add a payload byte to get correct CRC
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.reset_buffer()
        self.write(C)
        R = self.read(7)                    # grab with header, type, length, and crc
        if R == bytearray(C):
            return True
        else: 
            return False
    
    def get_fields(self,fields, ws = False): 
        '''Executes 380 GF command for an array of fields.  GF Command get current Temporary setting of 380
        ''' 
        # Take unit out of stream mode
        self.set_quiet()     
        num_fields = len(fields)
        C = [0x55, 0x55, ord('G'), ord('F'), num_fields * 2 + 1, num_fields]
        for field in fields:
            field_msb = (field & 0xFF00)  >> 8
            field_lsb = field & 0x00FF  
            C.insert(len(C), field_msb)
            C.insert(len(C), field_lsb)
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        R = self.read(num_fields * 4 + 1 + 7)
        data = []
        if R and R[0] == 85 and R[1] == 85:
            packet_crc = 256 * R[-2] + R[-1]                   # crc is last two bytes
            calc_crc = self.calc_crc(R[2:R[4]+5])
            if packet_crc == calc_crc:
                self.packet_type = '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
                data = self.parse_packet(R[5:R[4]+5], ws)
        return data
    
    def read_fields(self,fields, ws = False): 
        '''Executes 380 RF command for an array of fields.  RF Command get current Permanent setting of 380
        ''' 
        # Take unit out of stream mode
        self.set_quiet()     
        num_fields = len(fields)
        C = [0x55, 0x55, ord('R'), ord('F'), num_fields * 2 + 1, num_fields]
        for field in fields:
            field_msb = (field & 0xFF00)  >> 8
            field_lsb = field & 0x00FF  
            C.insert(len(C), field_msb)
            C.insert(len(C), field_lsb)
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        R = self.read(num_fields * 4 + 1 + 7)
        data = []
        if len(R) and R[0] == 85 and R[1] == 85:
            packet_crc = 256 * R[-2] + R[-1]                   # crc is last two bytes
            calc_crc = self.calc_crc(R[2:R[4]+5])
            if packet_crc == calc_crc:
                self.packet_type = '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
                data = self.parse_packet(R[5:R[4]+5], ws)
        return data
    
    def write_fields(self, field_value_pairs, ws=False):
        '''Executes 380 WF command for an array of fields, value pairs.  WF Command set Permanent setting for fields on 380
        '''
        self.set_quiet()     
        num_fields = len(field_value_pairs)
        C = [0x55, 0x55, ord('W'), ord('F'), num_fields * 4 + 1 , num_fields]
        FIELD = 0
        VALUE = 1
        for field_value in field_value_pairs:
            field_msb = (field_value[FIELD] & 0xFF00)  >> 8
            field_lsb = field_value[FIELD] & 0x00FF  
            if isinstance(field_value[VALUE], int):
                value_msb = (field_value[VALUE] & 0xFF00) >> 8
                value_lsb = field_value[VALUE] & 0x0FF
            elif isinstance(field_value[VALUE], str):
                value_msb = ord(field_value[VALUE][0])
                value_lsb = ord(field_value[VALUE][1])
            C.insert(len(C), field_msb)
            C.insert(len(C), field_lsb)
            C.insert(len(C), value_msb)
            C.insert(len(C), value_lsb)
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        time.sleep(1.0)
        R = self.read(num_fields * 2 + 1 +7)
        print(R)
        data = []
        if R[0] == 85 and R[1] == 85:
            packet_crc = 256 * R[-2] + R[-1]                   # crc is last two bytes
            if self.calc_crc(R[2:R[4]+5]) == packet_crc:
                if R[2] == 0 and R[3] == 0:
                    print('SET FIELD ERROR/FAILURE')
                    return
                else: 
                    self.packet_type =  '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
                    data = self.parse_packet(R[5:R[4]+5], ws)
        return data
    
    def set_fields(self, field_value_pairs, ws=False):
        '''Executes 380 SF command for an array of fields, value pairs.  SF Command sets Temporary setting for fields on 380
        '''
        self.set_quiet()     
        num_fields = len(field_value_pairs)
        C = [0x55, 0x55, ord('S'), ord('F'), num_fields * 4 + 1 , num_fields]
        FIELD = 0
        VALUE = 1
        for field_value in field_value_pairs:
            if (field_value[FIELD] == 1):
                self.odr_setting = field_value[VALUE]
            field_msb = (field_value[FIELD] & 0xFF00)  >> 8
            field_lsb = field_value[FIELD] & 0x00FF  
            value_msb = (field_value[VALUE] & 0xFF00) >> 8
            value_lsb = field_value[VALUE] & 0x0FF
            C.insert(len(C), field_msb)
            C.insert(len(C), field_lsb)
            C.insert(len(C), value_msb)
            C.insert(len(C), value_lsb)
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        R = self.read(num_fields * 2 + 1 +7)
        data = []
        if R[0] == 85 and R[1] == 85:
            packet_crc = 256 * R[-2] + R[-1]                   # crc is last two bytes
            if self.calc_crc(R[2:R[4]+5]) == packet_crc:
                if R[2] == 0 and R[3] == 0:
                    print('SET FIELD ERROR/FAILURE')
                    return
                else: 
                    self.packet_type =  '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
                    data = self.parse_packet(R[5:R[4]+5], ws)
        return data

    def set_quiet(self):
        '''Force 380 device to quiet / polled mode and inject 0.1 second delay, then clear input buffer
        '''
        self.stream_mode = 0 
        time.sleep(0.1) # wait for any packets to clear
        C = [0x55, 0x55, ord('S'), ord('F'), 0x05 , 0x01, 0x00, 0x01, 0x00, 0x00]
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.reset_buffer()
        self.write(C)
        self.read(10)
        time.sleep(0.1) # wait for command to take effect
        self.reset_buffer()
    
    def restore_odr(self):
        '''Restores device to odr mode vs SF command
        '''
        print('restore odr to ' + '{0:d}'.format(self.odr_setting))
        C = [0x55, 0x55, ord('S'), ord('F'), 0x05 , 0x01, 0x00, 0x01, 0x00, self.odr_setting]
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.reset_buffer()
        self.write(C)
        self.read(10)
        time.sleep(0.1) # wait for command to take effect
        self.reset_buffer()
        self.synced = 0
        self.packet_size = 0        
        self.packet_type = 0  
        self.elapsed_time_sec = 0
        self.data = {}
        self.stream_mode = 1
          
    def connect(self):
        '''Continous data collection loop to get and process data packets
        '''
        self.find_device()

        if self.odr_setting:
            self.restore_odr()
        else:
            print('no odr setting can connect')
            return
        while self.odr_setting and self.connected:
            if self.stream_mode:
                self.get_packet()
            else:
                time.sleep(0.05)
      
    def disconnect(self):
        '''Ends data collection loop.  Reset settings
        '''
        self.connected = 0
        self.device_id = 0
        self.odr_setting = 0
        self.stream_mode = 0
        self.synced = 0
        self.packet_size = 0        
        self.packet_type = 0        
      
    def get_packet(self):
        '''Syncs unit and gets packet.  Assumes unit is in stream_mode'''

        # Already synced
        if self.synced == 1:    
            # Read next packet of data based on expected packet size     
            S = self.read(self.packet_size + 7)
            
            if len(S) < 2:
                # Read Failed
                self.synced = 0                    
                return
            if S[0] == 85 and S[1] == 85:
                packet_crc = 256 * S[-2] + S[-1]    
                # Compare computed and read crc               
                if self.calc_crc(S[2:S[4]+5]) == packet_crc: 
                    # 5 is offset of first payload byte, S[4]+5 is offset of last payload byte     
                    self.data = self.parse_packet(S[5:S[4]+5])     
            else: 
                # Get synced and then read next packet
                self.sync()
                self.get_packet()
        else:
            # Get synced and then read next packet
            self.sync()
            self.get_packet()

    def sync(self,prev_byte = 0,bytes_read = 0):
        '''Syncs a 380 in Continuous / Stream mode.  Assumes longest packet is 40 bytes
            TODO: check this assumption
            TODO: add check of CRC
            :returns:
                true if synced, false if not
        '''
        S = self.read(1)
      
        if not S:
            return False
        if S[0] == 85 and prev_byte == 85:      # VALID HEADER FOUND
            # Once header is found then read off the rest of packet
            print('Synced!')
            self.synced = 1
            config_bytes = self.read(3)
            self.packet_type = '{0:1c}'.format(config_bytes[0]) + '{0:1c}'.format(config_bytes[1])
            self.packet_size = config_bytes[2]
            self.read(config_bytes[2] + 2)      # clear bytes off port, payload + 2 byte CRC
            return True
        else: 
            # Repeat sync to search next byte pair for header
            if bytes_read == 0:
                print('Connecting ....')
            bytes_read = bytes_read + 1
            print(bytes_read)
            self.synced = 0
            if (bytes_read < 40):
                self.sync(S[0], bytes_read)
            else:
                return False
    
    def start_bootloader(self):
        '''Starts bootloader
            :returns:
                True if bootloader mode entered, False if failed
        '''
        self.set_quiet()
        C = [0x55, 0x55, ord('J'), ord('I'), 0x00 ]
        crc = self.calc_crc(C[2:4] + [0x00])    # for some reason must add a payload byte to get correct CRC
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        time.sleep(2)   # must wait for boot loader to be ready
        R = self.read(5)
        if R[0] == 85 and R[1] == 85:
            self.packet_type =  '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
            if self.packet_type == 'JI':
                self.read(R[4]+2)
                print('bootloader ready')
                time.sleep(2)
                self.reset_buffer()
                if self.boot == 0:
                    print('resync with device')
                    time.sleep(2)
#                    self.find_device()
#                    self.reset_buffer()
                return True
            else: 
                return False
        else:
            return False
    
    def start_app(self):
        '''Starts app
        '''
        self.set_quiet()
        C = [0x55, 0x55, ord('J'), ord('A'), 0x00 ]
        crc = self.calc_crc(C[2:4] + [0x00])    # for some reason must add a payload byte to get correct CRC
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        time.sleep(1)
        R = self.read(7)    
        if R[0] == 85 and R[1] == 85:
            self.packet_type =  '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
            print(self.packet_type)

    def write_block(self, buf, data_len, addr):
        '''Executed WA command to write a block of new app code into memory
        '''
        print(data_len, addr)
        C = [0x55, 0x55, ord('W'), ord('A'), data_len+5]
        addr_3 = (addr & 0xFF000000) >> 24
        addr_2 = (addr & 0x00FF0000) >> 16
        addr_1 = (addr & 0x0000FF00) >> 8
        addr_0 = (addr & 0x000000FF)
        C.insert(len(C), addr_3)
        C.insert(len(C), addr_2)
        C.insert(len(C), addr_1)
        C.insert(len(C), addr_0)
        C.insert(len(C), data_len)
        for i in range(data_len):
            C.insert(len(C), ord(buf[i]))
        crc = self.calc_crc(C[2:C[4]+5])  
        crc_msb = int((crc & 0xFF00) >> 8)
        crc_lsb = int((crc & 0x00FF))
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        status = 0
        while (status == 0):
            self.write(C)
            if addr == 0:
               time.sleep(8)
            else:
               time.sleep(0.1)
            R = self.read(12)  #longer response
            if len(R) > 1 and R[0] == 85 and R[1] == 85:
                self.packet_type =  '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
                print(self.packet_type)
                if self.packet_type == 'WA':
                    status = 1
                else:
                    sys.exit()
                    print('retry 1')
                    status = 0
            else:
                print(len(R))
                print(R)
                self.reset_buffer()
                time.sleep(1)
                print('no packet')
                sys.exit()
        
    def upgrade_fw(self,file):
        '''Upgrades firmware of connected 380 device to file provided in argument
        '''
        print('upgrade fw')
        max_data_len = 240
        write_len = 0
        fw = open(self.filename, 'rb').read()
        fs_len = len(fw)

#        if not self.start_bootloader():
#            print('Bootloader Start Failed')
#            return False
       
        time.sleep(1)
        while (write_len < fs_len):
            packet_data_len = max_data_len if (fs_len - write_len) > max_data_len else (fs_len-write_len)
            # From IMUView 
            # Array.Copy(buf,write_len,write_buf,0,packet_data_len);
            write_buf = fw[write_len:(write_len+packet_data_len)]
            self.write_block(write_buf, packet_data_len, write_len)
            write_len += packet_data_len
        time.sleep(1)
        # Start new app
        self.start_app()
    
    def get_id_str(self):
        ''' Executes GP command and requests ID data from 380
            :returns:
                id string of connected device, or false if failed
        '''
        self.set_quiet()
        C = [0x55, 0x55, ord('G'), ord('P'), 0x02, ord('I'), ord('D') ]
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        R = self.read(5)    
        if len(R) and R[0] == 85 and R[1] == 85:
            self.packet_type =  '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
            payload_length = R[4]
            R = self.read(payload_length+2)
            id_str = self.parse_packet(R[0:payload_length])
            self.ID = id_str
            if id_str.find('load') == -1:
				self.boot = 0
            else:
                self.boot = 1			
            return id_str
        else: 
            return False

    def get_bit_status(self):
        ''' Executes GP command and requests bit stsatus from 380
            :returns:
        '''
        self.set_quiet()
        C = [0x55, 0x55, ord('G'), ord('P'), 0x02, ord('T'), ord('0') ]
        crc = self.calc_crc(C[2:C[4]+5])   
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        C.insert(len(C), crc_msb)
        C.insert(len(C), crc_lsb)
        self.write(C)
        R = self.read(5)    
        if len(R) and R[0] == 85 and R[1] == 85:
            self.packet_type =  '{0:1c}'.format(R[2]) + '{0:1c}'.format(R[3])
            payload_length = R[4]
            R = self.read(payload_length+2)
            id_str = self.parse_packet(R[0:payload_length])
            return id_str
        else: 
            return False

    def parse_packet(self, payload, ws = False):
        '''Parses packet payload to engineering units based on packet type
           Currently supports S0, S1, A1 packets.  Logs data if logging is on.
           Prints data if a GF/RF/SF/WF. Add A2, N0, N1 packet types.
        '''
        if self.packet_type == 'S0':
            '''S0 Payload Contents
                Byte Offset	Name	Format	Scaling	Units	Description
                0	xAccel	    I2	20/2^16	G	X accelerometer
                2	yAccel	    I2	20/2^16	G	Y accelerometer
                4	zAccel	    I2	20/2^16	G	Z accelerometer
                6	xRate   	I2	7*pi/2^16 [1260 deg/2^16]	rad/s [deg/sec]	X angular rate
                8	yRate	    I2	7*pi/2^16 [1260 deg/2^16]	rad/s [deg/sec]	Y angular rate
                10	zRate	    I2	7*pi/2^16 [1260 deg/2^16]	rad/s [deg/sec]	Z angular rate
                12	xMag	    I2	2/2^16	Gauss	X magnetometer
                14	yMag	    I2	2/2^16	Gauss	Y magnetometer
                16	zMag	    I2	2/2^16	Gauss	Z magnetometer
                18	xRateTemp	I2	200/2^16	deg. C	X rate temperature
                20	yRateTemp	I2	200/2^16	deg. C	Y rate temperature
                22	zRateTemp	I2	200/2^16	deg. C	Z rate temperature
                24	boardTemp	I2	200/2^16	deg. C	CPU board temperature
                26	GPSITOW	    U2	truncated	Ms	GPS ITOW (lower 2 bytes)
                28	BITstatus   U2 Master BIT and Status'''
            
            accels = [0 for x in range(3)] 
            for i in range(3):
                accel_int16 = (256 * payload[2*i] + payload[2*i+1]) - 65535 if 256 * payload[2*i] + payload[2*i+1] > 32767  else  256 * payload[2*i] + payload[2*i+1]
                accels[i] = (9.80665 * 20 * accel_int16) / math.pow(2,16)
  
            gyros = [0 for x in range(3)] 
            for i in range(3):
                gyro_int16 = (256 * payload[2*i+6] + payload[2*i+7]) - 65535 if 256 * payload[2*i+6] + payload[2*i+7] > 32767  else  256 * payload[2*i+6] + payload[2*i+7]
                gyros[i] = (1260 * gyro_int16) / math.pow(2,16) 

            mags = [0 for x in range(3)] 
            for i in range(3):
                mag_int16 = (256 * payload[2*i+12] + payload[2*i+13]) - 65535 if 256 * payload[2*i+12] + payload[2*i+13] > 32767  else  256 * payload[2*i+12] + payload[2*i+13]
                mags[i] = (2 * mag_int16) / math.pow(2,16) 

            temps = [0 for x in range(4)] 
            for i in range(4):
                temp_int16 = (256 * payload[2*i+18] + payload[2*i+19]) - 65535 if 256 * payload[2*i+18] + payload[2*i+19] > 32767  else  256 * payload[2*i+18] + payload[2*i+19]
                temps[i] = (200 * temp_int16) / math.pow(2,16)
        
            # Counter Value
            itow = 256 * payload[26] + payload[27]   

            # BIT Value
            bit = 256 * payload[28] + payload[29]   

            
            if self.data:
                prev_time = self.data['GPSITOW']
                if (itow > prev_time):
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (itow - prev_time)  
                else:
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (65535 - prev_time) + ( 1.0 / 65535.0 ) * itow 

            data = collections.OrderedDict([('time', self.elapsed_time_sec), ( 'xAccel', accels[0]), ('yAccel', accels[1]), ('zAccel', accels[2]), ('xRate', gyros[0]), \
                     ('yRate' , gyros[1]), ('zRate', gyros[2]), ('xMag', mags[0]), ('yMag', mags[1]), ('zMag', mags[2]), ('xRateTemp', temps[0]), \
                     ('yRateTemp', temps[1]), ('zRateTemp', temps[2]), ('boardTemp', temps[3]), ('GPSITOW', itow), ('BITstatus', bit )])

            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data

        elif self.packet_type == 'F1':
            '''F1 Payload Contents
                Byte Offset	Name	Format	Scaling	Units	Description
                0       payload length  U2                      user-definied payload length 
                2       payload content U1                      JSON file''' 

            # payload length
            len = 256 * payload[0] + payload[1]         

            if len == 0:
               return 0

            f = open("user_format.json", "r")

            f.close()

            return len 

        elif self.packet_type == 'S1':
            '''S1 Payload Contents
                Byte Offset	Name	Format	Scaling	Units	Description
                0	xAccel	I2	20/2^16	G	X accelerometer
                2	yAccel	I2	20/2^16	G	Y accelerometer
                4	zAccel	I2	20/2^16	G	Z accelerometer
                6	xRate	I2	7*pi/2^16   [1260 deg/2^16]	rad/s [deg/sec]	X angular rate
                8	yRate	I2	7*pi/2^16   [1260 deg/2^16]	rad/s [deg/sec]	Y angular rate
                10	zRate	I2	7*pi/2^16   [1260 deg/2^16]	rad/s [deg/sec]	Z angular rate
                12	xRateTemp	I2	200/2^16	deg. C	X rate temperature
                14	yRateTemp	I2	200/2^16	deg. C	Y rate temperature
                16	zRateTemp	I2	200/2^16	deg. C	Z rate temperature
                18	boardTemp	I2	200/2^16	deg. C	CPU board temperature
                20	counter         U2	-	packets	Output time stamp 
                22	BITstatus	U2	-	-	Master BIT and Status'''

            accels = [0 for x in range(3)] 
            for i in range(3):
                accel_int16 = (256 * payload[2*i] + payload[2*i+1]) - 65535 if 256 * payload[2*i] + payload[2*i+1] > 32767  else  256 * payload[2*i] + payload[2*i+1]
                accels[i] = (9.80665 * 20 * accel_int16) / math.pow(2,16)
 
            gyros = [0 for x in range(3)] 
            for i in range(3):
                gyro_int16 = (256 * payload[2*i+6] + payload[2*i+7]) - 65535 if 256 * payload[2*i+6] + payload[2*i+7] > 32767  else  256 * payload[2*i+6] + payload[2*i+7]
                gyros[i] = (1260 * gyro_int16) / math.pow(2,16) 

            temps = [0 for x in range(4)] 
            for i in range(4):
                temp_int16 = (256 * payload[2*i+12] + payload[2*i+13]) - 65535 if 256 * payload[2*i+12] + payload[2*i+13] > 32767  else  256 * payload[2*i+12] + payload[2*i+13]
                temps[i] = (200 * temp_int16) / math.pow(2,16)
        
            # Counter Value
            count = 256 * payload[20] + payload[21]   

            # BIT Value
            bit = 256 * payload[22] + payload[23]         

            if self.data:
                prev_time = self.data['counter']
                if (count > prev_time):
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (count - prev_time)  
                else:
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (65535 - prev_time) + ( 1.0 / 65535.0 ) * count 

            data = collections.OrderedDict([('time', self.elapsed_time_sec), ( 'xAccel', accels[0]), ('yAccel', accels[1]), ('zAccel', accels[2]), ('xRate', gyros[0]), \
                     ('yRate' , gyros[1]), ('zRate', gyros[2]), ('xRateTemp', temps[0]), \
                     ('yRateTemp', temps[1]), ('zRateTemp', temps[2]), ('boardTemp', temps[3]), ('counter', count), ('BITstatus', bit )])


            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data
      
        elif self.packet_type == 'A1': 
            '''A1 Payload Contents
                0	rollAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Roll angle
                2	pitchAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Pitch angle
                4	yawAngleMag	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Yaw angle (magnetic north)
                6	xRateCorrected	I2	7*pi/2^16[1260 deg/2^16]	rad/s  [deg/sec]	X angular rate Corrected
                8	yRateCorrected	I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Y angular rate Corrected
                10	zRateCorrected	I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Z angular rate Corrected
                12	xAccel	I2	20/2^16	g	X accelerometer
                14	yAccel	I2	20/2^16	g	Y accelerometer
                16	zAccel	I2	20/2^16	g	Z accelerometer
                18	xMag	I2	2/2^16	Gauss	X magnetometer
                20	yMag	I2	2/2^16	Gauss	Y magnetometer
                22	zMag	I2	2/2^16	Gauss	Z magnetometer
                24	xRateTemp	I2	200/2^16	Deg C	X rate temperature
                26	timeITOW	U4	1	ms	DMU ITOW (sync to GPS)
                30	BITstatus	U2	-	-	Master BIT and Status'''

            angles = [0 for x in range(3)] 
            for i in range(3):
                angle_int16 = (256 * payload[2*i] + payload[2*i+1]) - 65535 if 256 * payload[2*i] + payload[2*i+1] > 32767  else  256 * payload[2*i] + payload[2*i+1]
                angles[i] = (360.0 * angle_int16) / math.pow(2,16) 

            gyros = [0 for x in range(3)] 
            for i in range(3):
                gyro_int16 = (256 * payload[2*i+6] + payload[2*i+7]) - 65535 if 256 * payload[2*i+6] + payload[2*i+7] > 32767  else  256 * payload[2*i+6] + payload[2*i+7]
                gyros[i] = (1260 * gyro_int16) / math.pow(2,16) 

            accels = [0 for x in range(3)] 
            for i in range(3):
                accel_int16 = (256 * payload[2*i+12] + payload[2*i+13]) - 65535 if 256 * payload[2*i+12] + payload[2*i+13] > 32767  else  256 * payload[2*i+12] + payload[2*i+13]
                accels[i] = (9.80665 * 20 * accel_int16) / math.pow(2,16)
            
            mags = [0 for x in range(3)] 
            for i in range(3):
                mag_int16 = (256 * payload[2*i+18] + payload[2*i+19]) - 65535 if 256 * payload[2*i+18] + payload[2*i+19] > 32767  else  256 * payload[2*i+18] + payload[2*i+19]
                mags[i] = (2 * mag_int16) / math.pow(2,16) 

  
            temp_int16 = (256 * payload[2*i+24] + payload[2*i+25]) - 65535 if 256 * payload[2*i+24] + payload[2*i+25] > 32767  else  256 * payload[2*i+24] + payload[2*i+25]
            temp = (200 * temp_int16) / math.pow(2,16)
        
            # Counter Value
            itow = 16777216 * payload[26] + 65536 * payload[27] + 256 * payload[28] + payload[29]   

            # BIT Value
            bit = 256 * payload[30] + payload[31]         

            if self.data:
                prev_time = self.data['timeITOW']
                if (itow > prev_time):
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (itow - prev_time)  
                else:
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (65535 - prev_time) + ( 1.0 / 65535.0 ) * itow 

            data = collections.OrderedDict([('time', self.elapsed_time_sec), ('rollAngle', angles[0]),('pitchAngle', angles[1]),('yawAngleMag', angles[2]), \
                    ('xRateCorrected' , gyros[0]), ('yRateCorrected' , gyros[1]), ('zRateCorrected', gyros[2]), \
                    ( 'xAccel', accels[0]), ('yAccel', accels[1]), ('zAccel', accels[2]), \
                    ( 'xMag', mags[0]), ('yMag', mags[1]), ('zMag', mags[2]), ('xRateTemp', temp), \
                    ('timeITOW', itow), ('BITstatus', bit )])


            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data

        elif self.packet_type == 'A2': 
            '''A2 Payload Contents
                0	rollAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Roll angle
                2	pitchAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Pitch angle
                4	yawAngleMag	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Yaw angle (magnetic north)
                6	xRateCorrected	I2	7*pi/2^16[1260 deg/2^16]	rad/s  [deg/sec]	X angular rate Corrected
                8	yRateCorrected	I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Y angular rate Corrected
                10	zRateCorrected	I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Z angular rate Corrected
                12	xAccel	  I2	20/2^16	g	X accelerometer
                14	yAccel	  I2	20/2^16	g	Y accelerometer
                16	zAccel	  I2	20/2^16	g	Z accelerometer
                18	xRateTemp I2	200/2^16	Deg.C   X rate temperature 
                20	yRatetemp I2	200/2^16	Deg.C	Y rate temperature 
                22	zRateTemp I2	200/2^16	Deg.C   Z rate temperature 
                24	timeITOW	U4	1	ms	DMU ITOW (sync to GPS)
                28	BITstatus	U2	-	-	Master BIT and Status'''

            angles = [0 for x in range(3)] 
            for i in range(3):
                angle_int16 = (256 * payload[2*i] + payload[2*i+1]) - 65535 if 256 * payload[2*i] + payload[2*i+1] > 32767  else  256 * payload[2*i] + payload[2*i+1]
                angles[i] = (360.0 * angle_int16) / math.pow(2,16) 

            gyros = [0 for x in range(3)] 
            for i in range(3):
                gyro_int16 = (256 * payload[2*i+6] + payload[2*i+7]) - 65535 if 256 * payload[2*i+6] + payload[2*i+7] > 32767  else  256 * payload[2*i+6] + payload[2*i+7]
                gyros[i] = (1260 * gyro_int16) / math.pow(2,16) 

            accels = [0 for x in range(3)] 
            for i in range(3):
                accel_int16 = (256 * payload[2*i+12] + payload[2*i+13]) - 65535 if 256 * payload[2*i+12] + payload[2*i+13] > 32767  else  256 * payload[2*i+12] + payload[2*i+13]
                accels[i] = (9.80665 * 20 * accel_int16) / math.pow(2,16)
            
            temp = [0 for x in range(3)] 
            for i in range(3):
                temp_int16 = (256 * payload[2*i+18] + payload[2*i+19]) - 65535 if 256 * payload[2*i+18] + payload[2*i+19] > 32767  else  256 * payload[2*i+18] + payload[2*i+19]
                temp[i] = (200 * temp_int16) / math.pow(2,16)
        
            # Counter Value
            itow = 16777216 * payload[24] + 65536 * payload[25] + 256 * payload[26] + payload[27]   

            # BIT Value
            bit = 256 * payload[28] + payload[29]         

            if self.data:
                prev_time = self.data['timeITOW']
                if (itow > prev_time):
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (itow - prev_time)  
                else:
                    self.elapsed_time_sec += ( 1.0 / 65535.0 ) * (65535 - prev_time) + ( 1.0 / 65535.0 ) * itow 

            data = collections.OrderedDict([('time', self.elapsed_time_sec), ('rollAngle', angles[0]),('pitchAngle', angles[1]),('yawAngleMag', angles[2]), \
                    ('xRateCorrected' , gyros[0]), ('yRateCorrected' , gyros[1]), ('zRateCorrected', gyros[2]), \
                    ( 'xAccel', accels[0]), ('yAccel', accels[1]), ('zAccel', accels[2]), \
                    ( 'xRateTemp', temp[0]), ('yRateTemp', temp[1]), ('zRateTemp', temp[2]), \
                    ('timeITOW', itow), ('BITstatus', bit )])


            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data

        elif self.packet_type == 'A3': 
            '''A3 Payload Contents
                0	rollAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Roll angle
                2	pitchAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Pitch angle
                4	yawAngleMag	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Yaw angle (magnetic north)
                6	xRateScaled     I2	7*pi/2^16[1260 deg/2^16]	rad/s  [deg/sec]	X angular rate Corrected
                8	yRateScaled     I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Y angular rate Corrected
                10	zRateScaled     I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Z angular rate Corrected
                12	xAccel	  I2	20/2^16	g	X accelerometer
                14	yAccel	  I2	20/2^16	g	Y accelerometer
                16	zAccel	  I2	20/2^16	g	Z accelerometer
                18	xRateTemp I2	200/2^16	Deg.C   X rate temperature 
                20	yRatetemp I2	200/2^16	Deg.C	Y rate temperature 
                22	zRateTemp I2	200/2^16	Deg.C   Z rate temperature 
                24	timeITOW	U4	1	ms	DMU ITOW (sync to GPS)
                28	BITstatus	U2	-	-	Master BIT and Status'''

            angles = [0 for x in range(3)] 
            for i in range(3):
                angle_int16 = (256 * payload[2*i] + payload[2*i+1]) - 65535 if 256 * payload[2*i] + payload[2*i+1] > 32767  else  256 * payload[2*i] + payload[2*i+1]
                angles[i] = (360.0 * angle_int16) / math.pow(2,16) 

            gyros = [0 for x in range(3)] 
            for i in range(3):
                gyro_int16 = (256 * payload[2*i+6] + payload[2*i+7]) - 65535 if 256 * payload[2*i+6] + payload[2*i+7] > 32767  else  256 * payload[2*i+6] + payload[2*i+7]
                gyros[i] = (1260 * gyro_int16) / math.pow(2,16) 

            accels = [0 for x in range(3)] 
            for i in range(3):
                accel_int16 = (256 * payload[2*i+12] + payload[2*i+13]) - 65535 if 256 * payload[2*i+12] + payload[2*i+13] > 32767  else  256 * payload[2*i+12] + payload[2*i+13]
                accels[i] = (9.80665 * 20 * accel_int16) / math.pow(2,16)
            
            temp = [0 for x in range(3)] 
            for i in range(3):
                temp_int16 = (256 * payload[2*i+18] + payload[2*i+19]) - 65535 if 256 * payload[2*i+18] + payload[2*i+19] > 32767  else  256 * payload[2*i+18] + payload[2*i+19]
                temp[i] = (200 * temp_int16) / math.pow(2,16)
        
            # Counter Value
            itow = 16777216 * payload[24] + 65536 * payload[25] + 256 * payload[26] + payload[27]   

            # BIT Value
            bit = 256 * payload[28] + payload[29]         

            data = collections.OrderedDict([('rollAngle', angles[0]),('pitchAngle', angles[1]),('yawAngleMag', angles[2]), \
                    ('xRateScaled' , gyros[0]), ('yRateScaled' , gyros[1]), ('zRateScaled', gyros[2]), \
                    ( 'xAccel', accels[0]), ('yAccel', accels[1]), ('zAccel', accels[2]), \
                    ( 'xRateTemp', temp[0]), ('yRateTemp', temp[1]), ('zRateTemp', temp[2]), \
                    ('timeITOW', itow), ('BITstatus', bit )])


            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data

        elif self.packet_type == 'N0': 
            '''N0 Payload Contents
                0	rollAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Roll angle
                2	pitchAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Pitch angle
                4	yawAngleMag	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Yaw angle (magnetic north)
                6	xRateCorrected  I2	7*pi/2^16[1260 deg/2^16]	rad/s  [deg/sec]	X angular rate Corrected
                8	yRateCorrected  I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Y angular rate Corrected
                10	zRateCorrected  I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Z angular rate Corrected
                12	nVel            I2	512/2^16	m/s        North velocity	
                14	eVel            I2	512/2^16	m/s	   East Velocity 
                16	dVel            I2	512/2^16	m/s	   Down velocity 
                18	longiTude       I4	2*pi/2^32	Radians    Longitude 
                22	latiTude        I4	2*pi/2^32     	Radians    Latitude 
                26	altiTude        I2	2^14/2^16	m          GPS altitude 
                28	iTOW	        U2	truncated	ms	   ITOW (lower 2 bytes)
                30	BITstatus	U2	-	        -	   Master BIT and Status'''

            angles = [0 for x in range(3)] 
            for i in range(3):
                angle_int16 = (256 * payload[2*i] + payload[2*i+1]) - 65535 if 256 * payload[2*i] + payload[2*i+1] > 32767  else  256 * payload[2*i] + payload[2*i+1]
                angles[i] = (360.0 * angle_int16) / math.pow(2,16) 

            gyros = [0 for x in range(3)] 
            for i in range(3):
                gyro_int16 = (256 * payload[2*i+6] + payload[2*i+7]) - 65535 if 256 * payload[2*i+6] + payload[2*i+7] > 32767  else  256 * payload[2*i+6] + payload[2*i+7]
                gyros[i] = (1260 * gyro_int16) / math.pow(2,16) 

            vel = [0 for x in range(3)] 
            for i in range(3):
                vel_int16 = (256 * payload[2*i+12] + payload[2*i+13]) - 65535 if 256 * payload[2*i+12] + payload[2*i+13] > 32767  else  256 * payload[2*i+12] + payload[2*i+13]
                vel[i] = (512 * vel_int16) / math.pow(2,16)
            
            tude = [0 for x in range(3)] 
            for i in range(2):
                tude_int32 = (16777216 * payload[2*i+18] + 65536 * payload[2*i+19] + 256 * payload[2*i+20] + payload[21]) - 2147483648 if 16777216 * payload[2*i+18] + 65536 *  payload[2*i+19] + 256 * payload[2*i+20] + payload[21] > 2147483647  else  16777216 * payload[2*i+18] + 65536 * payload[2*i+19] + 256 * payload[2*i+20] + payload[21]
                tude[i] = (360.0  * tude_int32) / math.pow(2,32)
       
            # altitude
            tude_int16 = 256 * payload[26] + payload[27];
            tude[2] = (16384 * tude_int16) / math.pow(2,16); 
           
            # Counter Value
            itow = 256 * payload[28] + payload[29]   

            # BIT Value
            bit = 256 * payload[30] + payload[31]         

            data = collections.OrderedDict([('rollAngle', angles[0]),('pitchAngle', angles[1]),('yawAngleMag', angles[2]), \
                    ('xRateCorrected' , gyros[0]), ('yRateCorrected' , gyros[1]), ('zRateCOrrected', gyros[2]), \
                    ( 'nVel', vel[0]), ('eVel', vel[1]), ('dVel', vel[2]), \
                    ( 'longitude', tude[0]), ('latitude', tude[1]), ('altitude', tude[2]), \
                    ('iTOW', itow), ('BITstatus', bit )])


            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data

        elif self.packet_type == 'N1': 
            '''N0 Payload Contents
                0	rollAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Roll angle
                2	pitchAngle	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Pitch angle
                4	yawAngleMag	I2	2*pi/2^16 [360 deg/2^16]	Radians [deg]	Yaw angle (magnetic north)
                6	xRateCorrected  I2	7*pi/2^16[1260 deg/2^16]	rad/s  [deg/sec]	X angular rate Corrected
                8	yRateCorrected  I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Y angular rate Corrected
                10	zRateCorrected  I2	7*pi/2^16 [1260 deg/2^16]	rad/s  [deg/sec]	Z angular rate Corrected
                12	xAccel	        I2	20/2^16	                        g	                X accelerometer
                14	yAccel	        I2	20/2^16	                        g	                Y accelerometer
                16	zAccel	        I2	20/2^16	                        g	                Z accelerometer
                18	nVel            I2	512/2^16	m/s        North velocity	
                20	eVel            I2	512/2^16	m/s	   East Velocity 
                22	dVel            I2	512/2^16	m/s	   Down velocity 
                24	longiTude       I4	2*pi/2^32	Radians    Longitude 
                28	latiTude        I4	2*pi/2^32     	Radians    Latitude 
                32	altiTude        I2	2^14/2^16	m          GPS altitude 
                34	xRateTemp	I2	200/2^16	Deg C	X rate temperature
                36	iTOW	        U2	truncated	ms	   ITOW (lower 2 bytes)'''

            angles = [0 for x in range(3)] 
            for i in range(3):
                angle_int16 = (256 * payload[2*i] + payload[2*i+1]) - 65535 if 256 * payload[2*i] + payload[2*i+1] > 32767  else  256 * payload[2*i] + payload[2*i+1]
                angles[i] = (360.0 * angle_int16) / math.pow(2,16) 

            gyros = [0 for x in range(3)] 
            for i in range(3):
                gyro_int16 = (256 * payload[2*i+6] + payload[2*i+7]) - 65535 if 256 * payload[2*i+6] + payload[2*i+7] > 32767  else  256 * payload[2*i+6] + payload[2*i+7]
                gyros[i] = (1260 * gyro_int16) / math.pow(2,16) 

            accels = [0 for x in range(3)] 
            for i in range(3):
                accel_int16 = (256 * payload[2*i+12] + payload[2*i+13]) - 65535 if 256 * payload[2*i+12] + payload[2*i+13] > 32767  else  256 * payload[2*i+12] + payload[2*i+13]
                accels[i] = (9.80665 * 20 * accel_int16) / math.pow(2,16)
            vel = [0 for x in range(3)] 
            for i in range(3):
                vel_int16 = (256 * payload[2*i+18] + payload[2*i+19]) - 65535 if 256 * payload[2*i+18] + payload[2*i+19] > 32767  else  256 * payload[2*i+18] + payload[2*i+19]
                vel[i] = (512 * vel_int16) / math.pow(2,16)
            
            tude = [0 for x in range(3)] 
            for i in range(2):
                tude_int32 = (16777216 * payload[2*i+24] + 65536 * payload[2*i+25] + 256 * payload[2*i+26] + payload[27]) - 2147483648 if 16777216 * payload[2*i+24] + 65536 *  payload[2*i+25] + 256 * payload[2*i+26] + payload[27] > 2147483647  else  16777216 * payload[2*i+24] + 65536 * payload[2*i+25] + 256 * payload[2*i+26] + payload[27]
                tude[i] = (360.0 * tude_int32) / math.pow(2,32)
       
            # altitude
            tude_int16 = 256 * payload[32] + payload[33];
            tude[2] = (16384 * tude_int16) / math.pow(2,16); 
           
            temp_int16 = (256 * payload[2*i+34] + payload[2*i+35]) - 65535 if 256 * payload[2*i+34] + payload[2*i+35] > 32767  else  256 * payload[2*i+34] + payload[2*i+35]
            temp = (200 * temp_int16) / math.pow(2,16)

            # Counter Value
            itow = 256 * payload[28] + payload[29]   

            data = collections.OrderedDict([('rollAngle', angles[0]),('pitchAngle', angles[1]),('yawAngleMag', angles[2]), \
                    ('xRateCorrected' , gyros[0]), ('yRateCorrected' , gyros[1]), ('zRateCOrrected', gyros[2]), \
                    ( 'xAccel', accels[0]), ('yAccel', accels[1]), ('zAccel', accels[2]), \
                    ( 'nVel', vel[0]), ('eVel', vel[1]), ('dVel', vel[2]), \
                    ( 'longitude', tude[0]), ('latitude', tude[1]), ('altitude', tude[2]), \
                    ('xRateTemp', temp), ('iTOW', itow)])


            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data
            
        elif self.packet_type == 'T0':
            '''T0 Payload Contents
                Byte Offset	Name	 Format	Scaling	Units	Description
                0	BitStatus        U2	                Master BIT and Status Field 
                2	hardwareBIT      U2	                Hardware BIT Field 
                4	hardwarePowerBIT U2	                Hardware Power BIT Field 
                6	hardwareEnvironmentalBIT U2             Hardware Environmental BIT Field	
                8	comBit           U2                     communication BIT Field
                10	comSerialABIT    U2	                Communication Serial A BIT Field
                12      comSerialBBIT    U2                     Communication Serial B BIT Field	
                14	softwareBIT      U2                     Software BIT Filed	
                16	softwareAlgorithmBIT U2	                Software Algorithm BIT Field
                18	softwareDataBIT  U2	                Software Data BIT Field
                20	hardwareStatus   U2	                Hardware Status Field
                22	comStatus        U2	                Communication Status Field
                24	softwareStatus   U2	                Software Status Field
                26	sensorStatus     U2                     Sensor Status Field'''
            
            # BIT Status 
            bitStatus = 256 * payload[0] + payload[1]         

            # hardware bit 
            hardwareBit= 256 * payload[2] + payload[3]         

            # hardware power bit 
            hardwarePowerBit= 256 * payload[4] + payload[5]         

            # hardware environmental bit 
            hardwareEnvironmentalBit= 256 * payload[6] + payload[7]         

            # com bit 
            comBit = 256 * payload[8] + payload[9]         

            # com serial A bit 
            comSerialABit = 256 * payload[10] + payload[11]         

            # com serial B bit 
            comSerialBBit = 256 * payload[12] + payload[13]         

            # software bit 
            softwareBit= 256 * payload[14] + payload[15]         

            # software algorithm bit 
            softwareAlgBit= 256 * payload[16] + payload[17]         

            # software data bit 
            softwareDataBit= 256 * payload[18] + payload[19]         

            # hardware status 
            hardwareStatus = 256 * payload[20] + payload[21]         

            # com status 
            comStatus = 256 * payload[22] + payload[23]         

            # software status 
            softwareStatus = 256 * payload[24] + payload[25]         

            # sensor status 
            sensorStatus = 256 * payload[26] + payload[27]         

            data = collections.OrderedDict([( 'bitStatus', bitStatus), ('hardware bit', hardwareBit), ('hardwarePowerBit', hardwarePowerBit), \
                   ('hardwareEnvironmentalBit', hardwareEnvironmentalBit), ('comBit', comBit), ('com Serial A Bit', comSerialABit), \
                   ('com Serial B Bit', comSerialBBit), ('software bit', softwareBit), ('software algorithm bit', softwareAlgBit), \
                   ('software data bit', softwareDataBit), ('hardware status', hardwareStatus), ('com status', comStatus), \
                   ('software status', softwareStatus), ('sensor status', sensorStatus)])


            if self.logging == 1 and self.logger is not None:
                self.logger.log(data, self.odr_setting) 
            
            return data

        elif self.packet_type == 'SF':
            n = payload[0]
            for i in range(n):
                if ws == False:
                    print('Set Field: 0x{0:02X}'.format(payload[i*2+1]) + '{0:02X}'.format(payload[i*2+2]))
                else:
                    return 1
        elif self.packet_type == 'WF':
            n = payload[0]
            data = [0] * n  #empty array
            for i in range(n):
                if ws == False:
                    print('Write Field: 0x{0:02X}'.format(payload[i*2+1]) + '{0:02X}'.format(payload[i*2+2])) 
                else:
                    return 1
        elif self.packet_type == 'RF':
            n = payload[0]
            data = [0] * n  #empty array
            for i in range(n):
                if ws == False:
                    print(( 'Read Field: 0x{0:02X}'.format(payload[i*4+1]) + '{0:02X}'.format(payload[i*4+2]) 
                    + ' set to: 0x{0:02X}{1:02X}'.format(payload[i*4+3],payload[i*4+4])  
                    + ' ({0:1c}{1:1c})'.format(payload[i*4+3],payload[i*4+4]) ))
                else:
                    data[i] = [256 * payload[i*4+1] + payload[i*4+2], 256 * payload[i*4+3] + payload[i*4+4]]
            return data

        elif self.packet_type == 'GF':
            n = payload[0]
            data = [0] * n  #empty array
            for i in range(n):
                # remap about odr because unit was forced to quiet mode during GF read
                if ((256 * payload[i*4+1] + payload[i*4+2]) == 1):
                    payload[i*4+3] = 0
                    payload[i*4+4] = self.odr_setting
                if ws == False:
                    print(( 'Get Field: 0x{0:02X}'.format(payload[i*4+1]) + '{0:02X}'.format(payload[i*4+2]) 
                    + ' set to: 0x{0:02X}{1:02X}'.format(payload[i*4+3],payload[i*4+4])  
                    + ' ({0:1c}{1:1c})'.format(payload[i*4+3],payload[i*4+4]) ))
                else:
                    data[i] = [256 * payload[i*4+1] + payload[i*4+2], 256 * payload[i*4+3] + payload[i*4+4]]
            return data
            
        elif self.packet_type == 'VR':
            '''this packet type is obsolete'''
            print('Version String: {0}.{1}.{2}.{3}.{4}'.format(*payload))
        elif self.packet_type == 'ID':
            sn = int(payload[0] << 24) + int(payload[1] << 16) + int(payload[2] << 8) + int(payload[3])
            print('ID String: {0} {1}'.format(sn,payload[4:].decode()))
            return '{0} {1}'.format(sn,payload[4:].decode())

    def calc_crc(self,payload):
        '''Calculates CRC per 380 manual
        '''
        crc = 0x1D0F
        for bytedata in payload:
           crc = crc^(bytedata << 8) 
           for i in range(0,8):
                if crc & 0x8000:
                    crc = (crc << 1)^0x1021
                else:
                    crc = crc << 1

        crc = crc & 0xffff
        return crc

    def open(self, port, baud):
        try:
            self.ser = serial.Serial(port, baud, timeout = 0.1)
        except (OSError, serial.SerialException):
            print('serial port open exception' + port)

    def close(self):
            self.ser.close()
    
    def read(self,n):
        bytes = []
        try: 
            bytes = self.ser.read(n)
        except:
        # except (OSError, serial.SerialException):
            self.disconnect()    # sets connected to 0, and other related parameters to initial values
            print('serial exception read') 
            self.connect() 
        if bytes and len(bytes):
            return bytearray(bytes)
        else:
            print('empty read') 
            return bytearray(bytes)
    
    def write(self,n):
        try: 
            self.ser.write(n)
        except:
        # except (OSError, serial.SerialException):
            self.disconnect()   # sets connected to 0, and other related parameters to initial values  
            print('serial exception write')
            self.connect() 

    def reset_buffer(self):
        try:
            self.ser.reset_input_buffer()
        except:
        #except (OSError, serial.SerialException):
            self.disconnect()   # sets connected to 0, and other related parameters to initial values
            print('serial exception reset')   
            self.connect() 

        

if __name__ == "__main__":
    unit = OpenIMU()
    unit.find_device()
    print('file name ' + unit.filename)
    unit.start_bootloader()
    if unit.connected != 0:
        if unit.boot != 1:
            unit.close()
            unit.open(unit.port, 57600)
        unit.upgrade_fw(unit.filename)
    



import time
import datetime
import json
import requests
import threading

# need to find something python3 compatible  
# import urllib2

from azure.storage.blob import AppendBlobService
from azure.storage.blob import ContentSettings
from azure.storage.blob import BlockBlobService

class OpenIMULog:
    
    def __init__(self, imu, user = False):
        '''Initialize and create a CSV file
        '''
        self.name = 'data-' + datetime.datetime.now().strftime('%Y_%m_%d_%H_%M_%S') + '.csv'
        if user:
            self.user = user
            if self.user['fileName'] == '':
                self.user['fileName'] = self.name
            else:
                # self.user['fileName'] += '.csv'
                timeStrSaveFile = time.strftime("_%Y_%m_%d_%H_%M_%S", time.localtime()) + '.csv'
                self.user['fileName'] += timeStrSaveFile
            self.file = open('data/' + self.user['fileName'], 'w')
        else:
            self.file = open('data/' + self.name, 'w')
        self.first_row = 0
        # decode converts out of byte array
        self.ws = imu.ws
        self.sn = imu.device_id.split(" ")[0]
        self.pn = imu.device_id.split(" ")[1]
        self.device_id = imu.device_id
        self.odr_setting = imu.odr_setting
        self.packet_type = imu.packet_type
        self.imu_properties = imu.imu_properties

    # Parse the data, read in from the unit, and generate a data file using
    #   the json properties file to create a header and specify the precision
    #   of the data in the resulting data file.
    def log(self, imu, data):
        #
        output_packet = next((x for x in imu.imu_properties['userMessages']['outputPackets'] if x['name'] == imu.packet_type), None)

        '''Write row of CSV file based on data received.  Uses dictionary keys for column titles
        '''
        if not self.first_row:
            self.first_row = 1

            # Loop through each item in the data dictionary and create a header from the json
            #   properties that correspond to the items in the dictionary
            labels = ''
            keyIdx = -1
            for key in data:
                keyIdx= keyIdx + 1
                '''dataStr = output_packet['payload'][keyIdx]['name'] + \
                          ' [' + \
                          output_packet['payload'][keyIdx]['unit'] + \
                          ']'''
                dataStr = output_packet['payload'][keyIdx]['name'] + "(" + output_packet['payload'][keyIdx]['unit'] + ")"
                labels = labels + '{0:s},'.format(dataStr)
            
            # Remove the comma at the end of the string and append a new-line character
            labels = labels[:-1]
            header = labels + '\n'
        else:
            self.first_row += 1
            header = ''


        # Loop through the items in the data dictionary and append to an output string
        #   (with precision based on the data type defined in the json properties file)
        str = ''
        keyIdx = -1
        for key in data:
            keyIdx= keyIdx + 1
            outputPcktType = output_packet['payload'][keyIdx]['type']

            if outputPcktType == 'uint32' or outputPcktType == 'int32' or \
               outputPcktType == 'uint16' or outputPcktType == 'int16' or \
               outputPcktType == 'uint64' or outputPcktType == 'int64':
                # integers and unsigned integers
                str += '{0:d},'.format(data[key])
            elif outputPcktType == 'double':
                # double
                str += '{0:15.12f},'.format(data[key])
            elif outputPcktType == 'float':
                # print(3) #key + str(2))
                str += '{0:12.8f},'.format(data[key])
            elif outputPcktType == 'uint8':
                # byte
                str += '{0:d},'.format(data[key])
            elif outputPcktType == 'uchar' or outputPcktType == 'char':
                # character
                str += '{:},'.format(data[key])
            else:
                # unknown
                print(0)
                str += '{0:3.5f},'.format(data[key])

        # 
        str = str[:-1]
        str = str + '\n'
        self.file.write(header+str)

    def write_to_azure(self):
        # check for internet 
        # if not self.internet_on(): 
        #    return False

        # record file to cloud
        # f = open("data/" + self.name,"r")
        f = open("data/" + self.user['fileName'], "r")
        text = f.read()
       



        self.block_blob_service = BlockBlobService(account_name='navview',
                                                    sas_token=self.sas_token, # account_key
                                                    protocol='http')
        self.block_blob_service.create_blob_from_text(container_name='data',
                                                    blob_name=self.name,
                                                    text=text,
                                                    content_settings=ContentSettings(content_type='text/plain'))
        # self.block_blob_service.create_blob_from_text(container_name='data-1000',
        #                                             blob_name=self.name,
        #                                             text=text,
        #                                             content_settings=ContentSettings(content_type='text/plain'))


        # record record to ansplatform
        self.record_to_ansplatform()
        
        
    def record_to_ansplatform(self):
        # data = { "pn" : self.pn, "sn": self.sn, "fileName" : self.user['fileName'],  "url" : self.name, "imuProperties" : json.dumps(self.imu_properties),
        #          "sampleRate" : self.odr_setting, "packetType" : self.packet_type, "userId" : self.user['id'] }
        data = { 
                    "type":"IMU", 
                    "fileName" : self.user['fileName'], 
                    "url" : self.name,
                    "userId" : self.user['id'], 
                    "model" : "IMU300",
                    "logInfo": {
                        "pn" : self.pn, 
                        "sn": self.sn, 
                        "sampleRate" : self.odr_setting, 
                        "packetType" : self.packet_type,
                        "appVersion" : self.imu_properties['appName'] if 'appName' in self.imu_properties else '',
                        "imuProperties" : json.dumps(self.imu_properties)
                    } 
                }
        # host_address='http://40.118.233.18:3000/'
        host_address='https://api.aceinna.com/'        
        url = host_address + "api/recordLogs/post" #"https://api.aceinna.com/api/datafiles/replaceOrCreate"
        data_json = json.dumps(data)
        headers = {'Content-type': 'application/json', 'Authorization' : self.user['access_token'] }
        response = requests.post(url, data=data_json, headers=headers)
        response = response.json()
       
        # clean up
        self.name = ''

        return  #ends thread

    def internet_on(self):
        try:
            urllib2.urlopen('https://ans-platform.azurewebsites.net', timeout=1)
            return True
        except urllib2.URLError as err: 
            return False


    def get_sas_token(self):
        # print('user token',self.user['access_token'])
        try:
            # host_address='http://40.118.233.18:3000/'
            host_address='https://api.aceinna.com/'
            url = host_address+"token/storagesas"
            headers = {'Content-type': 'application/json', 'Authorization':  self.user['access_token']}
            response = requests.post(url, headers=headers)
            rev = response.json()
            if 'token' in rev:
                self.sas_token = rev['token']
            else:
                self.sas_token = ''
                print('Error: Get sas token failed!')
        except Exception as e:
            print('Exception when get_sas_token:', e)

    def close(self):
        time.sleep(0.1)
        if self.ws:
            self.file.close()
            self.get_sas_token()
            threading.Thread(target=self.write_to_azure).start()
        else:
            self.file.close()



import json 
import struct
import sys

class BootloaderInputPacket:

    def __init__(self, imu_properties, name, data_len = False, addr = False, data = False):
        
        self.name = name
        self.imu_properties = imu_properties

        # get name byte pair
        name_bytes =  list(struct.unpack('BB', bytearray(name, 'utf-8')))

        if not data_len and not addr and not data:
            S = [0x55, 0x55] + name_bytes + [0]  
            self.bytes = S + self.calc_crc(S[2:4] + [0x00])  
        else:
            payload = self.block_payload(data_len, addr, data)
            S = [0x55, 0x55] + name_bytes + [data_len+5] + payload
            self.bytes = S + self.calc_crc(S[2:S[4]+5])   

    def block_payload(self, data_len, addr, data):
        C = []
        addr_3 = (addr & 0xFF000000) >> 24
        addr_2 = (addr & 0x00FF0000) >> 16
        addr_1 = (addr & 0x0000FF00) >> 8
        addr_0 = (addr & 0x000000FF)
        C.insert(len(C), addr_3)
        C.insert(len(C), addr_2)
        C.insert(len(C), addr_1)
        C.insert(len(C), addr_0)
        C.insert(len(C), data_len)
        for i in range(data_len):
            if (sys.version_info > (3, 0)):
                C.insert(len(C), data[i])
            else:
                C.insert(len(C), ord(data[i]))
        return C
    
    def calc_crc(self,payload):
        '''Calculates 16-bit CRC-CCITT
        '''
        crc = 0x1D0F
        for bytedata in payload:
           crc = crc^(bytedata << 8) 
           for i in range(0,8):
                if crc & 0x8000:
                    crc = (crc << 1)^0x1021
                else:
                    crc = crc << 1

        crc = crc & 0xffff
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        return [ crc_msb, crc_lsb ]


"""
Global variables, imu, cli etc. are declared in the file
Created on 07-19-2018
"""

from openimu import OpenIMU



def get_app_urls():
    app_urls = ['https://raw.githubusercontent.com/Aceinna/python-openimu/master/app_config/Compass/openimu.json',
	        'https://raw.githubusercontent.com/Aceinna/python-openimu/master/app_config/IMU/openimu.json',
	        'https://raw.githubusercontent.com/Aceinna/python-openimu/master/app_config/INS/openimu.json',
	        'https://raw.githubusercontent.com/Aceinna/python-openimu/master/app_config/Leveler/openimu.json',
	        'https://raw.githubusercontent.com/Aceinna/python-openimu/master/app_config/OpenIMU/openimu.json',
	        'https://raw.githubusercontent.com/Aceinna/python-openimu/master/app_config/VG_AHRS/openimu.json'
	        ]
    return app_urls



def get_app_names():       
    app_names = ['Compass',
	        'IMU',
	        'INS',
	        'Leveler',
	        'OpenIMU',
	        'VG_AHRS'
	]        
    return app_names

app_str = ['INS','VG_AHRS','IMU','Compass','Leveler']

string_folder_path = 'app_config/APP_TYP/openimu.json'


imu = OpenIMU(ws=True)



# Copyright (C) 2018 Aceinna Navigation System Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance with the License. 
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, 
# software distributed under the License is distributed on an "AS IS" 
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express 
# or implied. See the License for the specific language governing 
# permissions and limitations under the License.

name = "python_openimu"




import json 
import struct

class InputPacket:

    def __init__(self, imu_properties, name, param = False, value = False):
        
        self.name = name
        self.imu_properties = imu_properties

        # get name byte pair
        name_bytes =  list(struct.unpack('BB', bytearray(name, 'utf-8')))

        if not param and not value:
            S = [0x55, 0x55] + name_bytes + [0]  
            self.bytes = S + self.calc_crc(S[2:4] + [0x00])  
        else:
            payload = self.unpack_payload(param, value)
            S = [0x55, 0x55] + name_bytes + [len(payload)] + payload
            self.bytes = S + self.calc_crc(S[2:S[4]+5])

    def unpack_payload(self, param = False, value = False):
        input_packet = next((x for x in self.imu_properties['userMessages']['inputPackets'] if x['name'] == self.name), None)
        if self.name == 'ma':
            input_action = next((x for x in input_packet['inputPayload'] if x['actionName'] == param), None)
            return [input_action['actionID']]
        elif input_packet != None:
            if input_packet['inputPayload']['type'] == 'paramId':
                return list(struct.unpack("4B", struct.pack("<L", param)))
            elif input_packet['inputPayload']['type'] == 'userParameter':
                payload = list(struct.unpack("4B", struct.pack("<L", param)))
                if self.imu_properties['userConfiguration'][param]['type'] == 'char8':
                    length = len(value)
                    payload += list(struct.unpack('{0}B'.format(length), bytearray(value,'utf-8')))
                    for i in range(8-length):
                        payload += [0x00]
                elif self.imu_properties['userConfiguration'][param]['type'] == 'int64':
                    payload += list(struct.unpack("8B", struct.pack("<q", value)))
                elif self.imu_properties['userConfiguration'][param]['type'] == 'double':
                    payload += list(struct.unpack("8B", struct.pack("<d", float(value))))
                return payload
    
    def calc_crc(self,payload):
        '''Calculates 16-bit CRC-CCITT
        '''
        crc = 0x1D0F
        for bytedata in payload:
           crc = crc^(bytedata << 8) 
           for i in range(0,8):
                if crc & 0x8000:
                    crc = (crc << 1)^0x1021
                else:
                    crc = crc << 1

        crc = crc & 0xffff
        crc_msb = (crc & 0xFF00) >> 8
        crc_lsb = (crc & 0x00FF)
        return [ crc_msb, crc_lsb ]


"""
Driver for Aceinna OpenIMU
Based on PySerial https://github.com/pyserial/pyserial
Created on 2018-06-17
@author: m5horton
"""

"""
WS Master Connection 
connect         - finds device, gets device_id/odr_setting, and loops
                - run this in thread otherwise blocking
disconnect      - ends loop

Device Discovery
find_device     - entry point to find a serial connected IMU
find_ports
autobaud

Logging
start_log
stop_log

Syncing 
sync            - syncs to a unit continuously transmitting, or a specific packet response

Data Functions

Serial          - a tiny layer on top of Pyserial to handle exceptions as means of device detection
open
close
read
write
reset_buffer
"""
import serial
import serial.tools.list_ports
import math
import string
import time
import sys
import os
from file_storage import OpenIMULog
import collections
import glob
import struct
import json
import threading
from pathlib import Path
from imu_input_packet import InputPacket
from bootloader_input_packet import BootloaderInputPacket
from azure.storage.blob import BlockBlobService
# import webbrowser # used for open ANS website by system browser
import requests
import binascii
import json
import global_vars as gl
 

class OpenIMU:
    def __init__(self, ws=False):
        '''Initialize and then start ports search and autobaud process
        '''
        self.ws = ws                # set to true if being run as a thread in a websocket server
        self.ser = None             # the active UART
        self.device_id = 0          # unit's id str
        self.odr_setting = 0        # units ODR rate
        self.paused = 1             # imu is successfully connected to a com port, kind of redundant with device_id property
        self.logging = 0            # logging on or off
        self.logger = None          # the file logger instance
        self.packet_size = 0        # expected size of packet 
        self.packet_type = 0        # expected type of packet
        self.data = {}              # placeholder imu measurements of last converted packeted
        self.data_buffer = []       # serial read buffer
        self.packet_buffer = []     # packet parsing buffer
        self.sync_state = 0
        self.sync_pattern = collections.deque(4*[0], 4)  # create 4 byte FIFO         

        #if no data folder, then creat one
        if not os.path.isdir("data"):
            print('creat data folder for store measure data in future')
            os.makedirs("data")
            
        if not os.path.exists('app_config'):
            print('downloading config json files from github, please waiting for a while')
            os.makedirs('app_config')
            for app_name in gl.get_app_names():
                os.makedirs('app_config'+ '/' + app_name)
            i = 0
            for url in gl.get_app_urls():
                filepath = 'app_config' + '/' + gl.get_app_names()[i] + '/' + 'openimu.json'
                i= i+1
                try:
                    r = requests.get(url) 
                    # r = requests.session().get(url)
                    with open(filepath, "wb") as code:
                        code.write(r.content)     
                except Exception as e:
                    print(e) 
        # else:
        #     print('load basic config json locally')
        
        # Load the basic openimu.json(IMU application)
        with open('app_config/IMU/openimu.json') as json_data:
            self.imu_properties = json.load(json_data)

    def find_device(self):
        ''' Finds active ports and then autobauds units
        '''        
        if self.try_last_port():
            self.set_connection_details()
            return True
        else:
            while not self.autobaud(self.find_ports()):
                time.sleep(0.1)
        return True
        
    def find_ports(self):
        ''' Lists serial port names. Code from
            https://stackoverflow.com/questions/12090503/listing-available-com-ports-with-python
            Successfully tested on Windows 8.1 x64, Windows 10 x64, Mac OS X 10.9.x / 10.10.x / 10.11.x and Ubuntu 14.04 / 14.10 / 15.04 / 15.10 with both Python 2 and Python 3.
            :raises EnvironmentError:
                On unsupported or unknown platforms
            :returns:
                A list of the serial ports available on the system
        '''
        print('scanning ports')

        #find system available ports
        portList = list(serial.tools.list_ports.comports())
        ports = [ p.device for p in portList]

        # if sys.platform.startswith('win'):
        #     #fond windows available ports already found
        #     portlist = list(serial.tools.list_ports.comports())
        #     ports = [ p.device for p in portlist]
            
        # elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        #     # this excludes your current terminal "/dev/tty"
        #     ports = glob.glob('/dev/tty[A-Za-z]*')
        # elif sys.platform.startswith('darwin'):
        #     ports = glob.glob('/dev/tty.*')
        # else:
        #     raise EnvironmentError('Unsupported platform')

        result = []
        for port in ports:
            if "Bluetooth" in port:
                continue
            else:
                print('Testing port ' + port)
                s = None
                try:
                    s = serial.Serial(port)
                    if s:
                        s.close()
                        result.append(port)
                # except:
                except Exception as e:
                    # print(e)                    
                    if sys.version_info[0] > 2:
                        if e.args[0].find('WindowsError') >= 0:
                            try:
                                if s:
                                    s.close()
                            except Exception as ee:
                                print (ee)
                                # self.disconnect()
                            
                    else:   
                        if e.message.find('Error') >= 0:
                            try:
                                if s:
                                    s.close()
                            except Exception as ee:
                                print (ee)
                                # self.disconnect()
                    pass
        return result

    def autobaud(self, ports):
        '''Autobauds unit - first check for stream_mode / continuous data, then check by polling unit
           :returns: 
                true when successful
        ''' 
        
        for port in ports:
            self.open(port)
            # TODO: change this to intelligently use openimu.json.  even save the last configuration 
            # user_configuration = self.imu_properties['userConfiguration']

            # ustrj = [x for x in self.imu_properties['userConfiguration'] if x['paramId'] == "2"]
            bandListFromOptions = []
            for x in self.imu_properties['userConfiguration']:
                if x['paramId'] == 2:
                    bandListFromOptions = x['options']
                    break
            for baud in bandListFromOptions:
                if self.ser:
                    self.ser.baudrate = baud
                    self.device_id = self.openimu_get_device_id()
                    if self.device_id:
                        print(self.device_id)
                    if self.device_id:
                        self.set_connection_details()
                        if sys.platform.startswith('win'):
                            self.ser.set_buffer_size(rx_size = 128000, tx_size = 128000)
                        return True
        return False
    
    def set_connection_details(self):
        if "Bootloader" in self.device_id:
            print('BOOTLOADER MODE') 
            print('Connected ....{0}'.format(self.device_id))
            print('Please Upgrade FW with upgrade_fw function')
        elif self.device_id:
            print('Connected ....{0}'.format(self.device_id))
        self.save_last_port()        
        # open the webside ans automatically by system browser        
        # webbrowser.open("http://40.118.233.18:8080/record", new=0, autoraise=True) 

    def try_last_port(self):
        connection = None
        portList = list(serial.tools.list_ports.comports())
        ports = [ p.device for p in portList]

        try:
            with open('app_config/connection.json') as json_data:
                connection = json.load(json_data)
            if connection:
                if not connection['port'] in ports:
                    return False
                print("try port saved in app_config/connection.json last time")
                self.open(port=connection['port'], baud=connection['baud'])
                if self.ser:
                    self.device_id = self.openimu_get_device_id()
                    if self.device_id:
                        print('autoconnected')
                        return True
                    else:
                        print('no port available in last recorded app_config/connection.json')
                        return False
                else:
                    return False
        except:
            return False

    def save_last_port(self):
        connection = { "port" : self.ser.port, "baud" : self.ser.baudrate }
        with open('app_config/connection.json', 'w') as outfile:
            json.dump(connection, outfile)
    
    def get_latest(self):
        '''Get latest converted IMU readings in converted units
            :returns:
                data object or error message for web socket server to pass to app
        '''
        if not self.paused:
            return self.data
        else: 
            return { 'error' : 'not streaming' }
    
    def start_log(self, data = False):
        '''Creates file or cloud logger.  Autostarts log activity if ws (websocket) set to false
        '''
        if self.ws == False: 
            if self.paused:
                self.connect()
            if self.odr_setting:
                if not os.path.exists("data"):
                    print("Create data/ ")
                    os.makedirs("data")
                print('Logging Started ...')
                self.logger = OpenIMULog(self,data)
                self.logging = 1
        elif not self.paused and self.odr_setting:
            self.logger = OpenIMULog(self,data)
            self.logging = 1
                
    def stop_log(self):
        '''Stops file or cloud logger
        '''
        self.pause()
        self.logging = 0
        self.logger.close()
        self.logger = None
        print('Logging Finished ...')

    def openimu_update_param(self, param, value):
        C = InputPacket(self.imu_properties, 'uP', param, value)
        self.write(C.bytes)
        #time.sleep(0.05)
        return self.openimu_get_packet('uP')  

    def openimu_get_param(self, param,):
        C = InputPacket(self.imu_properties, 'gP', param)
        self.write(C.bytes)
        #time.sleep(0.05)
        return self.openimu_get_packet('gP')

    def magneticAlignCmd(self,action):
        C = InputPacket(self.imu_properties, 'ma', action)
        self.write(C.bytes)

        if action == 'stored':
            time.sleep(1)
            returnedStatus = self.read(10000)

            decodedStatus = binascii.hexlify(returnedStatus)
            return self.decodeOutput(decodedStatus)

        if action == 'status':
            returnedStatus = self.ser.readline().strip()
            if sys.version_info[0] < 3:
                decodedStatus = binascii.hexlify(returnedStatus)
            else:
                decodedStatus = str(binascii.hexlify(returnedStatus), 'utf-8')

            # print (decodedStatus)
            if 'f12e' in decodedStatus:
                return 1

    def decodeOutput(self, value):
        hard_iron_x = dict()
        hard_iron_y = dict()
        soft_iron_ratio = dict()
        soft_iron_angle = dict()

        # output['hardIronX'] = self.hardIronCal(value[10:14], 'axis')
        # output['hardIronY'] = self.hardIronCal(value[14:18], 'axis')
        # output['SoftIronRatio'] = self.hardIronCal(value[18:22], 'ratio')
        # output['SoftIronAngle'] = self.hardIronCal(value[22:26], 'angle')

        hard_iron_x['value'] = self.hardIronCal(value[26:30], 'axis')
        hard_iron_x['name'] = 'Hard Iron X'
        hard_iron_x['argument'] = 'hard_iron_x'

        hard_iron_y['value'] = self.hardIronCal(value[30:34], 'axis')
        hard_iron_y['name'] = 'Hard Iron Y'
        hard_iron_y['argument'] = 'hard_iron_y'

        soft_iron_ratio['value'] = self.hardIronCal(value[34:38], 'ratio')
        soft_iron_ratio['name'] = 'Soft Iron Ratio'
        soft_iron_ratio['argument'] = 'soft_iron_ratio'

        soft_iron_angle['value'] = self.hardIronCal(value[38:42], 'angle')
        soft_iron_angle['name'] = 'Soft Iron Angle'
        soft_iron_angle['argument'] = 'soft_iron_angle'

        # output['hard_iron_y'] = self.hardIronCal(value[30:34], 'axis')
        # output['soft_iron_ratio'] = self.hardIronCal(value[34:38], 'ratio')
        # output['soft_iron_angle'] = self.hardIronCal(value[38:42], 'angle')

        output = [hard_iron_x,hard_iron_y,soft_iron_ratio,soft_iron_angle]

        return output

    def hardIronCal(self, value, type):
        decodedValue = int(value, 16)
        # print (decodedValue)
        if type == 'axis':
            if decodedValue > 2 ** 15:
                newDecodedValue = (decodedValue - 2 ** 16)
                return newDecodedValue / float(2 ** 15) * 8
            else:
                return decodedValue / float(2 ** 15) * 8

        if type == 'ratio':
            return decodedValue / float(2 ** 16 - 1)

        if type == 'angle':
            if decodedValue > 2 ** 15:
                decodedValue = decodedValue - 2 ** 16
                piValue = 2 ** 15 / math.pi
                return decodedValue / piValue

            piValue = 2 ** 15 / math.pi
            return decodedValue / piValue

    def openimu_save_config(self):
        C = InputPacket(self.imu_properties, 'sC')
        self.write(C.bytes)
        # this message currently does not return anything

    def openimu_get_all_param(self):
        C = InputPacket(self.imu_properties, 'gA')
        self.write(C.bytes)
        #time.sleep(0.05)
        return self.openimu_get_packet('gA')  

    def openimu_get_device_id(self):
        C = InputPacket(self.imu_properties, 'pG')
        self.write(C.bytes)
        #time.sleep(0.05)
        device_id = self.openimu_get_packet('pG')
        if device_id:
            # print('git device id')
            device_id = device_id.decode()
            return device_id
        else: 
            return False

    def openimu_get_user_app_id(self): 
        C = InputPacket(self.imu_properties, 'gV')
        # print('get id--------------1')
        self.write(C.bytes)        
        #time.sleep(0.05)
        return self.openimu_get_packet('gV')
    
    def openimu_version_compare(self, fw_version, js_version):
        '''app string should be matched with the corresponding json file.
        '''
        if js_version in fw_version:
            pass
            # print('fw & json version match')
        else:
            pass
            # print('fw & json version mismatch')

    def connect(self):
        '''Continous data collection loop to get and process data packets
        '''
        if not self.device_id:
            self.find_device()
        packet_rate = next((x for x in self.imu_properties['userConfiguration'] if x['name'] == 'Packet Rate'), None)
        odr_param = self.openimu_get_param(packet_rate['paramId'])
        try:
            self.odr_setting = odr_param['value']
        except:
            self.ord_setting = 0
        packet_type = next((x for x in self.imu_properties['userConfiguration'] if x['name'] == 'Packet Type'), None)
        packet_type = self.openimu_get_param(packet_type['paramId'])
        try: 
            self.packet_type = packet_type['value'][0:2]    # Set the Actual High Rate Packet
        except:
            self.packet_type = 0
        self.paused = 0
        # print("connect,  3333333333333333333333333333333")
        threading.Thread(target=self.start_collection_task).start()
            
    def pause(self):
        ''' Will End the data collection task and thread
        '''
        self.paused = 1 
        time.sleep(0.1)
        self.reset_buffer()
        
    def disconnect(self):
        '''Ends data collection loop and Reset settings
        '''
        if(self.logging == 1 ):
            self.stop_log() 
        self.pause()
        self.close()
        self.data = {}
        self.device_id = 0
        self.odr_setting = 0
        self.packet_size = 0        
        self.packet_type = 0
        
    def parse_payload(self, ws = False):
        '''Parses packet payload using openimu.json as reference
        '''
        payload = self.packet_buffer[3:self.packet_buffer[2]+3]   # extract the payload
        data = [] 

        # Find the packet in the imu_properties from unit's JSON description
        output_packet = next((x for x in self.imu_properties['userMessages']['outputPackets'] if x['name'] == self.packet_type), None)
        input_packet = next((x for x in self.imu_properties['userMessages']['inputPackets'] if x['name'] == self.packet_type), None)
        bootloader_packet = next((x for x in self.imu_properties['bootloaderMessages'] if x['name'] == self.packet_type), None)
      
        if output_packet != None:
            self.data = self.openimu_unpack_output_packet(output_packet, payload)
            if self.logging == 1 and self.logger is not None:
                self.logger.log(self, self.data)
                # print(self.data['lat'])
              
        elif input_packet != None:
            
            data = self.openimu_unpack_input_packet(input_packet['responsePayload'], payload)

        elif bootloader_packet != None:

            data = self.openimu_unpack_bootloader_packet(bootloader_packet['responsePayload'], payload)

        return data

    def calc_crc(self,payload):
        '''Calculates CRC per 380 manual
        '''
        crc = 0x1D0F
        for bytedata in payload:
           crc = crc^(bytedata << 8) 
           for i in range(0,8):
                if crc & 0x8000:
                    crc = (crc << 1)^0x1021
                else:
                    crc = crc << 1

        crc = crc & 0xffff
        return crc

    def open(self, port = False, baud = 57600):
        # simple open
        try:
            self.ser = serial.Serial(port, baud, timeout = 0.005)
        except Exception as e:
            print('serial port open exception: ' + port)
            self.ser = False

    def close(self):
        if self.ser:
            self.ser.reset_input_buffer()
            try: 
                self.ser.close()
            except:
                self.ser = False
    
    def read(self,n):
        bytes = []
        if self.ser:
            try: 
                bytes = self.ser.read(n)
                return bytearray(bytes)
            except serial.SerialException as e:
                if 'returned no data' in str(e):
                    return bytearray(bytes)
                else:
                    self.disconnect()    # sets connected to 0, and other related parameters to initial values
                    print('serial exception read') 
                    self.find_device() 
        
    def write(self,n):
        try: 
            self.ser.write(n)
        except Exception as e:
            print(e)
            self.disconnect()   # sets connected to 0, and other related parameters to initial values  
            print('serial exception write')
            self.find_device() 

    def reset_buffer(self):
        try:
            self.ser.reset_input_buffer()
        except:
            self.ser = False

    def openimu_unpack_output_packet(self, output_message, payload):
        length = 0
        pack_fmt = '<'
        for value in output_message['payload']:
            if value['type'] == 'float':
                pack_fmt += 'f'
                length += 4
            elif value['type'] == 'uint32':
                pack_fmt += 'I'
                length += 4
            elif value['type'] == 'int32':
                pack_fmt += 'i'
                length += 4
            elif value['type'] == 'int16':
                pack_fmt += 'h'
                length += 2
            elif value['type'] == 'uint16':
                pack_fmt += 'H'
                length += 2
            elif value['type'] == 'double':
                pack_fmt += 'd'
                length += 8
            elif value['type'] == 'int64':
                pack_fmt += 'q'
                length += 8
            elif value['type'] == 'uint64':
                pack_fmt += 'Q'
                length += 8
            elif value['type'] == 'char':
                pack_fmt += 'c'
                length += 1
            elif value['type'] == 'uchar':
                pack_fmt += 'B'
                length += 1
            elif value['type'] == 'uint8':
                pack_fmt += 'B'
                length += 1
        len_fmt = '{0}B'.format(length)
        try:
            b = struct.pack(len_fmt, *payload)
            data = struct.unpack(pack_fmt, b)
            out = [(value['name'],data[idx]) for idx,value in enumerate(output_message['payload'])]
            data = collections.OrderedDict(out)
            return data
        except Exception as e:
            print("error happened when decode the payload, pls restart IMU firmware: {0}".format(e))    
    
    def openimu_unpack_input_packet(self, input_message, payload):
        if input_message['type'] == 'userConfiguration':
            user_configuration = self.imu_properties['userConfiguration']
            params = []
            for parameter in user_configuration:
                id = parameter['paramId']
                type = parameter['type']
                name = parameter['name']
                value = self.openimu_unpack_one(type, payload[id*8:(id+1)*8])
                # print('{0}: {1}'.format(name,value))
                params.append({ "paramId": id, "name": name, "value": value})
            return params
        elif input_message['type'] == 'userParameter':
            user_configuration = self.imu_properties['userConfiguration']
            param_id = self.openimu_unpack_one('uint32', payload[0:4]) 
            if param_id >= len(user_configuration):
                return False
            param = user_configuration[param_id]
            param_value = self.openimu_unpack_one(param['type'], payload[4:12])
            # print('{0}: {1}'.format(param['name'], param_value))
            return { "paramId": param_id, "name": param['name'], "value": param_value }
        elif input_message['type'] == 'paramId':
            user_configuration = self.imu_properties['userConfiguration']
            error = self.openimu_unpack_one('uint32', payload[0:4]) 
            if not error:
                print("Successfully Updated")
            return { "error": error }
        elif input_message['type'] == 'string':
            text_string = self.openimu_unpack_one('string', payload)
            return text_string

    def openimu_unpack_bootloader_packet(self, bootloader_message, payload):
        if bootloader_message['type'] == 'ack':
            print('Success')
            return { "error": 0 }

    def openimu_unpack_one(self, type, data):
        if type == 'uint64':
            try:
                b = struct.pack('8B', *data)
            except:
                return False
            return struct.unpack('<Q', b)[0]
        elif type == 'int64':
            try:
                b = struct.pack('8B', *data)
            except:
                return False
            return struct.unpack('<q', b)[0]
        elif type == 'uint32':
            try: 
                b = struct.pack('4B', *data)
            except:
                return False
            return struct.unpack('<L', b)[0]
        elif type == 'char8':
            try:
                b = struct.pack('8B', *data)
                # print('unpack one---------0000')
                return b.decode()    
            except:
                return False 
        elif type == 'string':
            try:
                fmt_str = '{0}B'.format(len(data))
                b = struct.pack(fmt_str, *data)
                return b
            except:
                return False
        elif type == 'double':
            try:
                b = struct.pack('8B', *data)
            except:
                return False
            return struct.unpack('d', b)[0]

    def openimu_start_bootloader(self):
        packet = BootloaderInputPacket(self.imu_properties, 'JI')
        self.write(packet.bytes)
        self.openimu_get_packet('JI')    
        self.ser.baudrate = 57600   # Force baud rate to 57600
        bootloader_id = self.openimu_get_device_id()
        print(bootloader_id)
        return True
          
    def openimu_start_app(self):
        '''Starts app
        '''
        packet = BootloaderInputPacket(self.imu_properties, 'JA')
        self.write(packet.bytes)
        print('Restarting app ...')
        time.sleep(5)
        self.disconnect()
        self.reset_buffer()
        self.close()
        self.find_device()      # Must go thru reconnect process because baud rate may have changed during firmware load
        return True

    def openimu_write_block(self, data_len, addr, data):
        print(data_len, addr)
        packet = BootloaderInputPacket(self.imu_properties, 'WA', data_len, addr, data)
        self.write(packet.bytes)
        if addr == 0:
            time.sleep(5)
        self.openimu_get_packet('WA')    
       
    def openimu_upgrade_fw_prepare(self, file):
        if self.ws == False:
            fw_file = Path(file)

            if fw_file.is_file():
                self.fw = open(file, 'rb').read()
            else:
                return False

        if not self.openimu_start_bootloader():
            print('Bootloader Start Failed')
            return False

        if self.ws == True:
            self.block_blob_service = BlockBlobService(account_name='navview',
                                                    account_key='+roYuNmQbtLvq2Tn227ELmb6s1hzavh0qVQwhLORkUpM0DN7gxFc4j+DF/rEla1EsTN2goHEA1J92moOM/lfxg==',
                                                    protocol='http')
            self.block_blob_service.get_blob_to_path('apps', file, file)
            self.fw = open(file, 'rb').read()

        print('upgrade fw: %s' % file)
        self.max_data_len = 240
        self.addr = 0
        self.fs_len = len(self.fw)

        return True

    def openimu_finish_upgrade_fw(self):
        return self.addr >= self.fs_len

    def openimu_upgrade_fw(self, file):
        '''Upgrades firmware of connected 380 device to file provided in argument
        '''
        packet_data_len = self.max_data_len if (self.fs_len - self.addr) > self.max_data_len else (self.fs_len - self.addr)
        data = self.fw[self.addr : (self.addr + packet_data_len)]
        self.openimu_write_block(packet_data_len, self.addr, data)
        self.addr += packet_data_len

    def start_collection_task(self):
        # print("start collect task:, paused ========= ", self.paused == 1)
        while self.odr_setting and not self.paused:
            if self.odr_setting:
                self.openimu_get_packet(self.packet_type, True)     # get packet in stream mode
        print('End Collection Task')
        return False  # End Thread

    def openimu_get_packet(self,packet_type, stream = False):
        if not packet_type:
            return False
        data = False
        trys = 0

        while not data and trys < 200:
            self.data_buffer = self.read(10000)

            if self.data_buffer:
                data = self.parse_buffer(packet_type, stream)
            trys += 1
        return data

    def parse_buffer(self, packet_type, stream = False):
        if (sys.version_info > (3, 0)) and not isinstance(packet_type, str):
            packet_type_0 = packet_type[0]  
            packet_type_1 = packet_type[1]
        else:
            packet_type_0 = ord(packet_type[0])
            packet_type_1 = ord(packet_type[1])
        for i,new_byte in enumerate(self.data_buffer):
            self.sync_pattern.append(new_byte)
            if list(self.sync_pattern) == [85, 85, packet_type_0, packet_type_1]:
                self.packet_buffer = [packet_type_0, packet_type_1]
                self.sync_state = 1
            elif self.sync_state == 1:
                self.packet_buffer.append(new_byte)
                if len(self.packet_buffer) == self.packet_buffer[2] + 5:
                    packet_crc = 256 * self.packet_buffer[-2] + self.packet_buffer[-1]    
                    if packet_crc == self.calc_crc(self.packet_buffer[:-2]):
                        if not isinstance(packet_type, str) and not isinstance(packet_type, unicode):
                            # print('parse buffer---------')
                            self.packet_type = bytearray(packet_type).decode()
                        else:
                            self.packet_type = packet_type
                        data = self.parse_payload()
                        if not stream:
                            return data
                        else:
                            self.packet_buffer = []
                            self.sync_state = 0
                    else:
                        self.sync_state = 0  # CRC did not match
                        # print("crc error***************************************")
         

 #####       

if __name__ == "__main__":
    imu = OpenIMU()
    imu.find_device()
    imu.openimu_get_all_param()
    imu.start_log()
    time.sleep(20)
    imu.stop_log()
  
	
    



# Copyright (C) 2018 Aceinna Navigation System Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS"
# BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.

from setuptools import find_packages, setup

install_requires = [
    "pyserial",
    "pathlib",
    "psutil",
    "azure-storage-blob",
    "tornado"
]

setup(
    name="openimu",
    version="0.0.7",
    author="Aceinna OpenIMU",
    author_email="info@aceinna.com",
    description="Aceinna Open Source Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Aceinna/python-openimu",
    license="Apache 2.0",
    python_requires=">=2.7, !=3.0.*, !=3.1.*",
    install_requires=install_requires,
    packages=['openimu'],
    package_data={'': ['LICENSE']},
    data_files=[('', ['openimu/app_config/IMU/openimu.json','openimu/app_config/INS/openimu.json','openimu/app_config/VG_AHRS/openimu.json','openimu/app_config/Leveler/openimu.json','openimu/app_config/Compass/openimu.json',])],
    include_package_data=True,
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    entry_points={
        "console_scripts": [
            "openimu = openimu.__main__:main",
        ]
    }
)



import sys
import time
if sys.version_info[0] > 2:
    from openimu.global_vars import imu
    from openimu.commands import OpenIMU_CLI
else:
    from global_vars import imu
    from commands import OpenIMU_CLI

def main(args=None):
    if args is None:
        args = sys.argv[1:]

    print("This is openimu main routine.")
    imu.find_device()
    cli = OpenIMU_CLI() 
    cli.command_handler()

if __name__ == "__main__":
    main()



import math

class Quat:
    def __init__(self):
        self.q = [1.0, 0.0, 0.0, 0.0]
        self.pq = [1.0, 0.0, 0.0, 0.0]

    def update_quat(self, w, t):
        rvec = [ w['wx'] * t, w['wy'] * t, w['wz'] * t ]   
        fetarad = math.sqrt(rvec[0] * rvec[0] + rvec[1] * rvec[1] + rvec[2] * rvec[2])
        fetarad2 = fetarad * fetarad

        if fetarad2 < 0.02:
            sinhalfeta = fetarad * (0.5 - 0.02083333333 * fetarad2)
        elif fetarad2 < 0.06:
            fetarad4 = fetarad2 * fetarad2
            sinhalfeta = fetarad * (0.5 - 0.02083333333 * fetarad2 + 0.0002604166667 * fetarad4)
        else:
            sinhalfeta = math.sin(0.5 * fetarad)


        if fetarad != 0.0:
            ftmp = sinhalfeta / fetarad
            self.pq[1] = rvec[0] * ftmp		
            self.pq[2] = rvec[1] * ftmp		
            self.pq[3] = rvec[2] * ftmp		
        else:
            self.pq[1] = 0.0
            self.pq[2] = 0.0
            self.pq[3] = 0.0

        fvecsq = self.pq[1] * self.pq[1] + self.pq[2] * self.pq[2] + self.pq[3] * self.pq[3]
	    
        if fvecsq <= 1.0:
            self.pq[0] = math.sqrt(1.0 - fvecsq)
        else: 
            self.pq[0] = 0.0

        tmpq = [0.0, 0.0, 0.0, 0.0 ]
        tmpq[0] = self.q[0]*self.pq[0] - self.q[1]*self.pq[1] - self.q[2]*self.pq[2] - self.q[3]*self.pq[3]
        tmpq[1] = self.q[0]*self.pq[1] + self.pq[0]*self.q[1] + self.q[2]*self.pq[3] - self.q[3]*self.pq[2]
        tmpq[2] = self.q[0]*self.pq[2] + self.pq[0]*self.q[2] + self.q[3]*self.pq[1] - self.q[1]*self.pq[3]
        tmpq[3] = self.q[0]*self.pq[3] + self.pq[0]*self.q[3] + self.q[1]*self.pq[2] - self.q[2]*self.pq[1]


        qMag = math.sqrt(tmpq[0] * tmpq[0] + tmpq[1] * tmpq[1] + tmpq[2] * tmpq[2] + tmpq[3] * tmpq[3])
        
        if qMag == 0.0:
            pass
        elif self.q[0] < 0.0:
            qMag = -1*qMag
        
        if qMag > 0.0: 
            self.q[0] = tmpq[0]/qMag
            self.q[1] = tmpq[1]/qMag
            self.q[2] = tmpq[2]/qMag        
            self.q[3] = tmpq[3]/qMag

    def to_matrix4(self):
        w = self.q[0]
        x = self.q[1]
        y = self.q[2]
        z = self.q[3]

        x2 = x + x
        y2 = y + y
        z2 = z + z        
        xx = x * x2
        xy = x * y2
        xz = x * z2
        yy = y * y2
        yz = y * z2
        zz = z * z2
        wx = w * x2
        wy = w * y2
        wz = w * z2

        m = [[0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, 0.0, 0.0],
             [0.0, 0.0, 0.0, 0.0]]

        m[0][0] = 1.0 - (yy + zz)
        m[0][1] = xy + wz
        m[0][2] = xz - wy
        m[0][3] = 0.0

        m[1][0] = xy - wz
        m[1][1] = 1.0 - (xx + zz)
        m[1][2] = yz + wx
        m[1][3] = 0.0

        m[2][0] = xz + wy
        m[2][1] = yz - wx
        m[2][2] = 1.0 - (xx + yy)
        m[2][3] = 0.0

        m[3][0] = 0.0
        m[3][1] = 0.0
        m[3][2] = 0.0
        m[3][3] = 1.0

        return m

    def print_rpy(self, mtx):
        thetaX = math.asin(mtx[1][2])
        thetaY = 0.0;
        thetaZ = 0.0;

        if thetaX < math.pi/2:
            if thetaX > -1 * math.pi/2:
                thetaZ = math.atan2(-mtx[1][0], mtx[1][1])
                thetaY = math.atan2(-mtx[0][2], mtx[2][2])
            else:
                thetaZ = -1 * math.atan2(mtx[2][0], mtx[0][0])
                thetaY = 0.0
        else:
            thetaZ = math.atan2(mtx[2][0], mtx[0][0])
            thetaY = 0.0

        hdgDegrees = 57.29 * thetaY
        pitchDegrees = 57.29 * thetaX
        rollDegrees = 57.29 * thetaZ

        print('{0:2.3f},{1:2.3f},{2:2.3f}'.format(hdgDegrees,pitchDegrees,rollDegrees))
    
    def print_euler(self):
        
        eroll  = 57.29 * math.atan2(2.0 * (self.q[3] * self.q[2] + self.q[0] * self.q[1]) , 1.0 - 2.0 * (self.q[1] * self.q[1] + self.q[2] * self.q[2]))
        epitch = 57.29 * math.asin(2.0 * (self.q[2] * self.q[0] - self.q[3] * self.q[1]))
        eyaw   = 57.29 * math.atan2(2.0 * (self.q[3] * self.q[0] + self.q[1] * self.q[2]) , - 1.0 + 2.0 * (self.q[0] * self.q[0] + self.q[1] * self.q[1]))

        print('{0:2.3f},{1:2.3f},{2:2.3f}'.format(eroll,epitch,eyaw))
